/* zenith_theme_draw.c - Zenith GTK Theme drawing functions.
** Copyright (C) 1999 Karl J. Ots.
** 
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
** 
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
*/

#include <math.h>
#include <gtk/gtk.h>

#include "zenith_theme.h"

/* Shortcuts for drawing horizontal and vetical lines */
#define gdk_draw_line_h(w,gc,x1,x2,y) (gdk_draw_line((w),(gc),(x1),(y),(x2),(y)))
#define gdk_draw_line_v(w,gc,y1,y2,x) (gdk_draw_line((w),(gc),(x),(y1),(x),(y2)))

/* Some debugging stuff */
static gchar buf[1024];
#define DETAIL  if (detail) { sprintf (buf, "Detail: %s\n", detail); g_print (buf); }

/* Some other useful macros */
#define IS_EVEN(x) (((x) & 1) == 0)
#define IS_ODD(x)  (((x) & 1) == 1)

static void
zenith_draw_hline      (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x1,
			gint x2,
			gint y)
{
  gint thickness_light, thickness_dark;
  gint i, n;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if (detail && !strcmp("menuitem", detail))
    y++;

  thickness_light = style->klass->ythickness / 2;
  thickness_dark = style->klass->ythickness - thickness_light;
  
  if (area)
  {
    gdk_gc_set_clip_rectangle(style->light_gc[state_type], area);
    gdk_gc_set_clip_rectangle(style->dark_gc[state_type],  area);
  }
  
  for (i = 0; i < thickness_dark; i++)
  {
    gdk_draw_line_h (window, style->light_gc[state_type], x2 - i - 1, x2, y + i);
    gdk_draw_line_h (window, style->dark_gc[state_type],  x1, x2 - i - 1, y + i);
  }
  
  y += thickness_dark;
  for (i = 0; i < thickness_light; i++)
  {
    gdk_draw_line_h (window, style->dark_gc[state_type],  x1, x1 + thickness_light - i - 1, y + i);
    gdk_draw_line_h (window, style->light_gc[state_type], x1 + thickness_light - i - 1, x2, y + i);
  }
  
  if (area)
  {
    gdk_gc_set_clip_rectangle(style->light_gc[state_type], NULL);
    gdk_gc_set_clip_rectangle(style->dark_gc[state_type],  NULL);
  }
}

static void
zenith_draw_vline      (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint y1,
			gint y2,
			gint x)
{
  gint thickness_light, thickness_dark;
  gint i;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  thickness_light = style->klass->xthickness / 2;
  thickness_dark = style->klass->xthickness - thickness_light;

  if (area)
  {
    gdk_gc_set_clip_rectangle(style->light_gc[state_type], area);
    gdk_gc_set_clip_rectangle(style->dark_gc[state_type],  area);
  }
  
  for (i = 0; i < thickness_dark; i++)
  {
    gdk_draw_line_v (window, style->light_gc[state_type], y2 - i - 1, y2, x + i);
    gdk_draw_line_v (window, style->dark_gc[state_type],  y1, y2 - i - 1, x + i);
  }
  
  x += thickness_dark;
  for (i = 0; i < thickness_light; i++)
  {
    gdk_draw_line_v (window, style->dark_gc[state_type],  y1, y1 + thickness_light - i, x + i);
    gdk_draw_line_v (window, style->light_gc[state_type], y1 + thickness_light - i, y2, x + i);
  }
  
  if (area)
  {
    gdk_gc_set_clip_rectangle(style->light_gc[state_type], NULL);
    gdk_gc_set_clip_rectangle(style->dark_gc[state_type],  NULL);
  }
}

static void
zenith_draw_shadow     (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height)
{
  GdkGC           *light_gc, *half_light_gc, *half_dark_gc, *dark_gc;
  GdkGC           *gc1, *gc2, *gc3, *gc4, *bg_gc;
  ZenithThemeData *zenith_theme_data;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  zenith_theme_data = style->engine_data;
  if (zenith_theme_data)
  {
    light_gc      = style->light_gc[state_type];
    half_light_gc = zenith_theme_data->half_light_gc[state_type];
    half_dark_gc  = zenith_theme_data->half_dark_gc[state_type];
    dark_gc       = style->dark_gc[state_type];
  }
  else
  {
    g_warning ("Zenith: zenith_draw_shadow: Theme data is NULL");
    light_gc      = style->white_gc;
    half_light_gc = style->light_gc[state_type];
    half_dark_gc  = style->dark_gc[state_type];
    dark_gc       = style->black_gc;
  }


  switch (shadow_type) {
  case GTK_SHADOW_OUT:
    gc1 = light_gc;
    gc2 = half_light_gc;
    gc3 = half_dark_gc;
    gc4 = dark_gc;
    break;
  
  case GTK_SHADOW_IN:
    gc1 = dark_gc;
    gc2 = half_dark_gc;
    gc3 = half_light_gc;
    gc4 = light_gc;
    break;
  
  case GTK_SHADOW_ETCHED_OUT:
    gc1 = light_gc;
    gc2 = dark_gc;
    gc3 = light_gc;
    gc4 = dark_gc;
    break;
  
  case GTK_SHADOW_ETCHED_IN:
    gc1 = dark_gc;
    gc2 = light_gc;
    gc3 = dark_gc;
    gc4 = light_gc;
    break;
  
  case GTK_SHADOW_NONE:
    return;
  }
  bg_gc = style->bg_gc[GTK_STATE_NORMAL];
    
  if (area)
  {
    gdk_gc_set_clip_rectangle (gc1, area);
    gdk_gc_set_clip_rectangle (gc2, area);
    gdk_gc_set_clip_rectangle (gc3, area);
    gdk_gc_set_clip_rectangle (gc4, area);
    gdk_gc_set_clip_rectangle (bg_gc, area);
  }

  width--;
  height--;
  
  gdk_draw_line_h (window, gc1, x + 1, x + width  - 1, y);
  gdk_draw_line_v (window, gc1, y + 1, y + height - 1, x);
  
  gdk_draw_line_h (window, gc2, x + 1, x + width  - 2, y + 1);
  gdk_draw_line_v (window, gc2, y + 1, y + height - 1, x + 1);
  
  gdk_draw_line_h (window, gc3, x + 2, x + width  - 1, y + height - 1);
  gdk_draw_line_v (window, gc3, y + 1, y + height - 1, x + width  - 1);
  
  gdk_draw_line_h (window, gc4, x + 1, x + width  - 1, y + height);
  gdk_draw_line_v (window, gc4, y + 1, y + height - 1, x + width);

  if (area)
  {
    gdk_gc_set_clip_rectangle (gc1, NULL);
    gdk_gc_set_clip_rectangle (gc2, NULL);
    gdk_gc_set_clip_rectangle (gc3, NULL);
    gdk_gc_set_clip_rectangle (gc4, NULL);
    gdk_gc_set_clip_rectangle (bg_gc, NULL);
  }
}

static void
zenith_draw_polygon    (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			GdkPoint * points,
			gint npoints,
			gint fill)
{
  /* This hav been copied verbatim from Redmond95 */
  
#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif /* M_PI */
#ifndef M_PI_4
#define M_PI_4  0.78539816339744830962
#endif /* M_PI_4 */

  static const gdouble pi_over_4 = M_PI_4;
  static const gdouble pi_3_over_4 = M_PI_4 * 3;

  GdkGC              *gc1;
  GdkGC              *gc2;
  GdkGC              *gc3;
  GdkGC              *gc4;
  gdouble             angle;
  gint                xadjust;
  gint                yadjust;
  gint                i;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);
  g_return_if_fail(points != NULL);

  switch (shadow_type)
  {
  case GTK_SHADOW_IN:
    gc1 = style->bg_gc[state_type];
    gc2 = style->dark_gc[state_type];
    gc3 = style->light_gc[state_type];
    gc4 = style->black_gc;
    break;
  case GTK_SHADOW_ETCHED_IN:
    gc1 = style->light_gc[state_type];
    gc2 = style->dark_gc[state_type];
    gc3 = style->dark_gc[state_type];
    gc4 = style->light_gc[state_type];
    break;
  case GTK_SHADOW_OUT:
    gc1 = style->dark_gc[state_type];
    gc2 = style->light_gc[state_type];
    gc3 = style->black_gc;
    gc4 = style->bg_gc[state_type];
    break;
  case GTK_SHADOW_ETCHED_OUT:
    gc1 = style->dark_gc[state_type];
    gc2 = style->light_gc[state_type];
    gc3 = style->light_gc[state_type];
    gc4 = style->dark_gc[state_type];
    break;
  default:
    return;
  }
  
  if (area)
  {
    gdk_gc_set_clip_rectangle(gc1, area);
    gdk_gc_set_clip_rectangle(gc2, area);
    gdk_gc_set_clip_rectangle(gc3, area);
    gdk_gc_set_clip_rectangle(gc4, area);
  }
  
  if (fill)
    gdk_draw_polygon(window, style->bg_gc[state_type], TRUE, points, npoints);
  
  npoints--;
  
  for (i = 0; i < npoints; i++)
  {
    if ((points[i].x == points[i + 1].x) &&
	(points[i].y == points[i + 1].y))
    {
      angle = 0;
    }
    else
    {
      angle = atan2(points[i + 1].y - points[i].y,
		    points[i + 1].x - points[i].x);
    }
    
    if ((angle > -pi_3_over_4) && (angle < pi_over_4))
    {
      if (angle > -pi_over_4)
      {
	xadjust = 0;
	yadjust = 1;
      }
      else
      {
	xadjust = 1;
	yadjust = 0;
      }
      
      gdk_draw_line(window, gc1,
		    points[i].x - xadjust, points[i].y - yadjust,
		    points[i + 1].x - xadjust, points[i + 1].y - yadjust);
      gdk_draw_line(window, gc3,
		    points[i].x, points[i].y,
		    points[i + 1].x, points[i + 1].y);
    }
    else
    {
      if ((angle < -pi_3_over_4) || (angle > pi_3_over_4))
      {
	xadjust = 0;
	yadjust = 1;
      }
      else
      {
	xadjust = 1;
	yadjust = 0;
      }
      
      gdk_draw_line(window, gc4,
		    points[i].x + xadjust, points[i].y + yadjust,
		    points[i + 1].x + xadjust, points[i + 1].y + yadjust);
      gdk_draw_line(window, gc2,
		    points[i].x, points[i].y,
		    points[i + 1].x, points[i + 1].y);
    }
  }
  if (area)
  {
    gdk_gc_set_clip_rectangle(gc1, NULL);
    gdk_gc_set_clip_rectangle(gc2, NULL);
    gdk_gc_set_clip_rectangle(gc3, NULL);
    gdk_gc_set_clip_rectangle(gc4, NULL);
  }
}

static void
zenith_draw_arrow      (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			GtkArrowType arrow_type,
			gint fill,
			gint x,
			gint y,
			gint width,
			gint height)
{
  gint half_width, half_height;
  gint b,i,d;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  /* The region the arrow is being drawn in needs to be square */
  /* and must have odd-length sides                            */
  width  = MIN (width, height);
  height = MIN (width, height);
  if (IS_EVEN (width)) { width--; height--; }

  /* Make sure the area is large enough to hold an arrow */
  if (width <= 7) return;

  if (detail && !strcmp ("menuitem", detail))
  {
    i = 0;
    d = 2;
    b = 1;
  }
  else
  {
    i = 0;
    d = 1;

    if      (width > 12) b = 4;
    else if (width > 8)  b = 3;
    else                 b = 2;
			   
    gtk_paint_box (style, window, state_type, shadow_type, area, widget, detail,
		   x, y, width, height);
  }
  x += b;
  y += b;
  width -= 2 * b + 1;
  height -= 2 * b + 1;
  half_width = width / 2;
  half_height = height / 2;

  if (area)
    gdk_gc_set_clip_rectangle (style->black_gc, area);
      
  switch (arrow_type) {
  case GTK_ARROW_UP:
    for (i = i; (half_width + i) <= width; i += d)
    {
      gdk_draw_line (window, style->black_gc, x, y + half_height + i, x + half_width, y + i);
      gdk_draw_line (window, style->black_gc, x + half_width, y + i, x + width, y + half_height + i);
    }
    break;
    
  case GTK_ARROW_DOWN:
    for (i = i; (half_width + i) <= width; i += d)
    {
      gdk_draw_line (window, style->black_gc, x, y + i, x + half_width, y + half_height + i);
      gdk_draw_line (window, style->black_gc, x + half_width, y + half_height + i, x + width, y + i);
    }
    break;
    
  case GTK_ARROW_LEFT:
    for (i = i; (half_width + i) <= width; i += d)
    {
      gdk_draw_line (window, style->black_gc, x + i, y + half_height, x + half_width + i, y);
      gdk_draw_line (window, style->black_gc, x + i, y + half_height, x + half_width + i, y + height);
    }
    break;
    
  case GTK_ARROW_RIGHT:
    for (i = i; (half_width + i) <= width; i += d)
    {
      gdk_draw_line (window, style->black_gc, x + i, y, x + half_width + i, y + half_height);
      gdk_draw_line (window, style->black_gc, x + i, y + height, x + half_width + i, y + half_height);
    }
    break;
  }
  
  if (area)
    gdk_gc_set_clip_rectangle (style->black_gc, NULL);
}

static void
zenith_draw_diamond    (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height)
{
  GdkGC           *light_gc, *half_light_gc, *half_dark_gc, *dark_gc;
  GdkGC           *bg_gc, *gc1, *gc2, *gc3, *gc4;
  GdkPoint         points[4];
  ZenithThemeData *zenith_theme_data;
  gint             half_width, half_height;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
  
  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  zenith_theme_data = style->engine_data;
  
  /* Width and height need to be odd */
  if (IS_EVEN (width))  width--;
  if (IS_EVEN (height)) height--;

  width--;
  height--;
  half_width  =  width / 2;
  half_height = height / 2;
  
  bg_gc = style->bg_gc[state_type];
  if (detail && !strcmp ("radiobutton", detail))
  {
    if (state_type != GTK_STATE_INSENSITIVE)
    {
      if (shadow_type == GTK_SHADOW_IN)
	bg_gc = style->base_gc[GTK_STATE_NORMAL];
      else
	bg_gc = style->bg_gc[GTK_STATE_NORMAL];
    }
      
    state_type = GTK_STATE_NORMAL;
    shadow_type = GTK_SHADOW_IN;
  }
  
  if (zenith_theme_data)
  {
    light_gc      = style->light_gc[state_type];
    half_light_gc = zenith_theme_data->half_light_gc[state_type];
    half_dark_gc  = zenith_theme_data->half_dark_gc[state_type];
    dark_gc       = style->dark_gc[state_type];
  }
  else
  {
    g_warning ("Zenith: zenith_draw_diamond: Theme data is NULL");
    light_gc      = style->white_gc;
    half_light_gc = style->light_gc[state_type];
    half_dark_gc  = style->dark_gc[state_type];
    dark_gc       = style->black_gc;
  }
  
  switch (shadow_type) {
  case GTK_SHADOW_OUT:
    gc1 = light_gc;
    gc2 = half_light_gc;
    gc3 = half_dark_gc;
    gc4 = dark_gc;
    break;
  
  case GTK_SHADOW_IN:
    gc1 = dark_gc;
    gc2 = half_dark_gc;
    gc3 = half_light_gc;
    gc4 = light_gc;
    break;
  
  case GTK_SHADOW_ETCHED_OUT:
    gc1 = light_gc;
    gc2 = dark_gc;
    gc3 = light_gc;
    gc4 = dark_gc;
    break;
  
  case GTK_SHADOW_ETCHED_IN:
    gc1 = dark_gc;
    gc2 = light_gc;
    gc3 = dark_gc;
    gc4 = light_gc;
    break;
    
  case GTK_SHADOW_NONE:
    return;
  }

  if (area)
  {
    gdk_gc_set_clip_rectangle (gc1, area);
    gdk_gc_set_clip_rectangle (gc2, area);
    gdk_gc_set_clip_rectangle (gc3, area);
    gdk_gc_set_clip_rectangle (gc4, area);
    gdk_gc_set_clip_rectangle (bg_gc, area);
  }
  
  points[0].x = x;
  points[0].y = y + half_height;
  points[1].x = x + half_width;
  points[1].y = y;
  points[2].x = x + width;
  points[2].y = y + half_height;
  points[3].x = x + half_width;
  points[3].y = y + height;
  gdk_draw_polygon (window, bg_gc, TRUE, points, 4);

  gdk_draw_line (window, gc1, x, y + half_height, x + half_width, y);
  gdk_draw_line (window, gc1, x + half_width, y, x + width - 1, y + half_height - 1);

  gdk_draw_line (window, gc2, x + 1, y + half_height, x + half_width, y + 1);
  gdk_draw_line (window, gc2, x + half_width, y + 1, x + width - 2, y + half_height - 1);

  gdk_draw_line (window, gc3, x + 2, y + half_height + 1, x + half_width, y + height - 1);
  gdk_draw_line (window, gc3, x + half_width, y + height - 1, x + width - 1, y + half_height);

  gdk_draw_line (window, gc4, x + 1, y + half_height + 1, x + half_width, y + height);
  gdk_draw_line (window, gc4, x + half_width, y + height, x + width, y + half_height);
  
  if (area)
  {
    gdk_gc_set_clip_rectangle (gc1, NULL);
    gdk_gc_set_clip_rectangle (gc2, NULL);
    gdk_gc_set_clip_rectangle (gc3, NULL);
    gdk_gc_set_clip_rectangle (gc4, NULL);
    gdk_gc_set_clip_rectangle (bg_gc, NULL);
  }
}

static void
zenith_draw_oval       (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height)
{
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
}

static void
zenith_draw_string     (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			const gchar * string)
{
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
  
  if (area)
  {
    gdk_gc_set_clip_rectangle (style->light_gc[GTK_STATE_INSENSITIVE], area);
    gdk_gc_set_clip_rectangle (style->fg_gc[state_type], area);
  }

  if (state_type == GTK_STATE_INSENSITIVE)
    gdk_draw_string (window, style->font, style->light_gc[GTK_STATE_INSENSITIVE], x + 1, y + 1, string);
  gdk_draw_string (window, style->font, style->fg_gc[state_type], x, y, string);
  
  if (area)
  {
    gdk_gc_set_clip_rectangle (style->light_gc[GTK_STATE_INSENSITIVE], NULL);
    gdk_gc_set_clip_rectangle (style->fg_gc[state_type], NULL);
  }
}

static void
zenith_draw_box        (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height)
{
  GdkGC *gc;
  gint   i,j,b;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);
  
  gc = style->bg_gc[state_type];
  b = 0;
  if (detail)
  {
    if (!strcmp("bar", detail))
    {
      gc = style->bg_gc[GTK_STATE_INSENSITIVE];
      shadow_type = GTK_SHADOW_NONE;
      b = 1;
    }
    else if (!strcmp("menuitem", detail))
    {
      gc = style->bg_gc[GTK_STATE_INSENSITIVE];
      shadow_type = GTK_SHADOW_IN;
      b = 1;
    }
    else if (!strcmp("trough", detail))
    {
      shadow_type = GTK_SHADOW_OUT;
    }
  }

  x += b;
  y += b;
  width -= 2 * b;
  height -= 2 * b;
  if ((width <= 0) || (height <= 0))
    return;
  
  if (!style->bg_pixmap[state_type] ||
      (gdk_window_get_type(window) == GDK_WINDOW_PIXMAP))
  {
    if (area)
      gdk_gc_set_clip_rectangle (gc, area);
    
    if (shadow_type != GTK_SHADOW_NONE)
      gdk_draw_rectangle (window, gc, TRUE, x + 1, y + 1, width - 2, height - 2);
    else
      gdk_draw_rectangle (window, gc, TRUE, x, y, width, height);

     if (area)
      gdk_gc_set_clip_rectangle (gc, NULL);
  }
  else
    gtk_style_apply_default_pixmap (style, window, state_type, area,
				    x, y, width, height);
  
  gtk_paint_shadow (style, window, state_type, shadow_type, area, widget, detail,
		    x, y, width, height);

  if (detail && !strcmp("slider", detail) && (shadow_type != GTK_SHADOW_NONE))
    gtk_paint_shadow (style, window, state_type, GTK_SHADOW_ETCHED_IN, area, widget, detail,
		      x + 4, y + 4, width - 8, height - 8);
}

static void
zenith_draw_flat_box   (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height)
{
  GdkGC *gc;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);
  
  gc = style->bg_gc[state_type];
  if (detail)
  {
    if (!strcmp ("viewportbin", detail))
      gc = style->bg_gc[GTK_STATE_NORMAL];
    else if (!strcmp ("entry_bg", detail) || !strcmp ("tooltip", detail))
      gc = style->base_gc[state_type];
    else if (!strcmp ("radiobutton", detail) || !strcmp ("checkbutton", detail))
      return;
  }

  if ((!style->bg_pixmap[state_type]) || (gc != style->bg_gc[state_type]) ||
      (gdk_window_get_type (window) == GDK_WINDOW_PIXMAP))
  {
    if (area)
    {
      gdk_gc_set_clip_rectangle (gc, area);
      gdk_gc_set_clip_rectangle (style->black_gc, area);
    }

    gdk_draw_rectangle (window, gc, TRUE, x, y, width, height);
    
    if (detail && !strcmp ("tooltip", detail))
      gdk_draw_rectangle (window, style->black_gc, FALSE, x, y, width - 1, height - 1);
    
    if (area)
    {
      gdk_gc_set_clip_rectangle (gc, NULL);
      gdk_gc_set_clip_rectangle (style->black_gc, NULL);
    }
  }
  else
    gtk_style_apply_default_pixmap (style, window, state_type, area,
				    x, y, width, height);
}

static void
zenith_draw_check      (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height)
{
  GdkGC *gc;
  gint   b;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  if (state_type == GTK_STATE_INSENSITIVE)
    gc = style->bg_gc[GTK_STATE_INSENSITIVE];
  else
    gc = style->base_gc[GTK_STATE_NORMAL];
  
  if (area)
  {
    gdk_gc_set_clip_rectangle (gc, area);
    gdk_gc_set_clip_rectangle (style->black_gc, area);
  }

  gdk_draw_rectangle (window, gc, TRUE, x + 1, y + 1, width - 2, height - 2);

  if ((shadow_type == GTK_SHADOW_IN) && (state_type != GTK_STATE_INSENSITIVE))
  {
    width--; height--;
    b = 4;
    
    gdk_draw_line (window, style->black_gc, x + b, y + b, x + width - b, y + height - b);
    gdk_draw_line (window, style->black_gc, x + b + 1, y + b, x + width - b, y + height - b - 1);
    gdk_draw_line (window, style->black_gc, x + b, y + b + 1, x + width - b - 1, y + height - b);

    gdk_draw_line (window, style->black_gc, x + b, y + height - b, x + width - b, y + b);
    gdk_draw_line (window, style->black_gc, x + b + 1, y + height - b, x + width - b, y + b + 1);
    gdk_draw_line (window, style->black_gc, x + b, y + height - b - 1, x + width - b - 1, y + b);
    
    width++; height++;
  }
  
  if (area)
  {
    gdk_gc_set_clip_rectangle (gc, NULL);
    gdk_gc_set_clip_rectangle (style->black_gc, NULL);
  }

  gtk_paint_shadow (style, window, GTK_STATE_NORMAL, GTK_SHADOW_IN, area, widget, detail,
		    x, y, width, height);
}

static void
zenith_draw_option     (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height)
{
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  gtk_paint_diamond (style, window, state_type, shadow_type, area, widget, detail,
		     x, y, width, height);
}

static void
zenith_draw_cross      (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height)
{
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
}

static void
zenith_draw_ramp       (GtkStyle * style,
		       GdkWindow * window,
		       GtkStateType state_type,
		       GtkShadowType shadow_type,
		       GdkRectangle * area,
		       GtkWidget * widget,
		       gchar * detail,
		       GtkArrowType arrow_type,
		       gint x,
		       gint y,
		       gint width,
		       gint height)
{
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
}

static void
zenith_draw_tab        (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height)
{
  GdkGC           *half_dark_gc;
  ZenithThemeData *zenith_theme_data;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  zenith_theme_data = style->engine_data;
  if (zenith_theme_data)
  {
    half_dark_gc  = zenith_theme_data->half_dark_gc[state_type];
  }
  else
  {
    g_warning ("Zenith: zenith_draw_box_gap: Theme data is NULL");
    half_dark_gc  = style->dark_gc[state_type];
  }
  
  if (detail && !strcmp("optionmenutab", detail))
  {
    if (x > 0) x--;
    if (y > 0) y--;

    if (area)
      gdk_gc_set_clip_rectangle (half_dark_gc, area);

    gdk_draw_rectangle (window, half_dark_gc, TRUE,
			x + 3, y + 3, width, height);
    
    if (area)
      gdk_gc_set_clip_rectangle (half_dark_gc, NULL);
  }

  gtk_paint_box (style, window, state_type, shadow_type, area, widget, detail,
		 x, y, width, height);
}

static void
zenith_draw_shadow_gap (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height,
			GtkPositionType gap_side,
			gint gap_x,
			gint gap_width)
{
  GdkRectangle rect;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);
  
  gtk_paint_shadow (style, window, state_type, shadow_type, area, widget, detail,
		    x, y, width, height);

  switch (gap_side) {
  case GTK_POS_TOP:
    rect.x      = x + gap_x;
    rect.y      = y;
    rect.width  = gap_width;
    rect.height = 2;
    break;
    
  case GTK_POS_BOTTOM:
    rect.x      = x + gap_x;
    rect.y      = y + height - 2;
    rect.width  = gap_width;
    rect.height = 2;
    break;
    
  case GTK_POS_LEFT:
    rect.x      = x;
    rect.y      = y + gap_x;
    rect.width  = 2;
    rect.height = gap_width;
    break;
    
  case GTK_POS_RIGHT:
    rect.x      = x + width - 2;
    rect.y      = y + gap_x;
    rect.width  = 2;
    rect.height = gap_width;
    break;
  }

  gtk_style_apply_default_pixmap (style, window, state_type, area,
  			  rect.x, rect.y, rect.width, rect.height);

}

static void
zenith_draw_box_gap    (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height,
			GtkPositionType gap_side,
			gint gap_x,
			gint gap_width)
{
  GdkGC           *light_gc, *half_light_gc, *half_dark_gc, *dark_gc;
  ZenithThemeData *zenith_theme_data;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
  
  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);
  
  zenith_theme_data = style->engine_data;
  if (zenith_theme_data)
  {
    light_gc      = style->light_gc[state_type];
    half_light_gc = zenith_theme_data->half_light_gc[state_type];
    half_dark_gc  = zenith_theme_data->half_dark_gc[state_type];
    dark_gc       = style->dark_gc[state_type];
  }
  else
  {
    g_warning ("Zenith: zenith_draw_box_gap: Theme data is NULL");
    light_gc      = style->white_gc;
    half_light_gc = style->light_gc[state_type];
    half_dark_gc  = style->dark_gc[state_type];
    dark_gc       = style->black_gc;
  }
  
  gtk_paint_box (style, window, state_type, shadow_type, area, widget, detail,
		 x, y, width, height);

  if (area)
  {
    gdk_gc_set_clip_rectangle (light_gc, area);
    gdk_gc_set_clip_rectangle (half_light_gc, area);
    gdk_gc_set_clip_rectangle (half_dark_gc,  area);
    gdk_gc_set_clip_rectangle (dark_gc, area);
  }
  
  switch (gap_side) {
  case GTK_POS_TOP:
    gtk_style_apply_default_pixmap (style, window, state_type, area,
				    x + gap_x, y, gap_width, 2);
    
    gap_width--; width--; height--;
    
    gdk_draw_point (window, light_gc, x + gap_x, y);
    gdk_draw_point (window, half_light_gc, x + gap_x + 1, y);
    if (gap_x == 0)
    {
      gdk_draw_point (window, light_gc, x + gap_x, y + 1);
      gdk_draw_point (window, half_light_gc, x + gap_x + 1, y + 1);
    }
    else
      gdk_draw_point (window, half_light_gc, x + gap_x, y + 1);

    gdk_draw_point (window, half_dark_gc,  x + gap_x + gap_width - 1, y);
    gdk_draw_point (window, dark_gc, x + gap_x + gap_width, y);

    if ((x + gap_x + gap_width) == width)
    {
      gdk_draw_point (window, half_dark_gc,  x + gap_x + gap_width - 1, y + 1);
      gdk_draw_point (window, dark_gc, x + gap_x + gap_width, y + 1);
    }
    else
      gdk_draw_point (window, half_light_gc, x + gap_x + gap_width, y + 1);
    break;
    
  case GTK_POS_BOTTOM:
    gtk_style_apply_default_pixmap (style, window, state_type, area,
				    x + gap_x, y + height - 2, gap_width, 2);
    
    gap_width--; width--; height--;
    
    gdk_draw_point (window, light_gc, x + gap_x, y + height);
    gdk_draw_point (window, half_light_gc, x + gap_x + 1, y + height);
    if (gap_x == 0)
    {
      gdk_draw_point (window, light_gc, x + gap_x, y + height - 1);
      gdk_draw_point (window, half_light_gc, x + gap_x + 1, y + height - 1);
    }
    else
      gdk_draw_point (window, half_dark_gc,  x + gap_x, y + height - 1);

    gdk_draw_point (window, half_dark_gc,  x + gap_x + gap_width - 1, y + height);
    gdk_draw_point (window, dark_gc, x + gap_x + gap_width, y + height);
    if ((x + gap_x + gap_width) == width)
    {
    gdk_draw_point (window, half_dark_gc,  x + gap_x + gap_width - 1, y + height - 1);
    gdk_draw_point (window, dark_gc, x + gap_x + gap_width, y + height - 1);
    }
    else
      gdk_draw_point (window, half_dark_gc,  x + gap_x + gap_width, y + height - 1);
    break;
    
  case GTK_POS_LEFT:
    gtk_style_apply_default_pixmap (style, window, state_type, area,
				    x, y + gap_x, 2, gap_width);
    
    gap_width--; width--; height--;

    gdk_draw_point (window, light_gc, x, y + gap_x);
    gdk_draw_point (window, half_light_gc, x, y + gap_x + 1);
    if (gap_x == 0)
    {
      gdk_draw_point (window, light_gc, x + 1, y + gap_x);
      gdk_draw_point (window, half_light_gc, x + 1, y + gap_x + 1);
    }
    else
      gdk_draw_point (window, half_light_gc, x + 1, y + gap_x);
    
    gdk_draw_point (window, half_dark_gc,  x, y + gap_x + gap_width - 1);
    if ((y + gap_x + gap_width) == height)
    {
      gdk_draw_point (window, dark_gc, x, y + gap_x + gap_width);
      gdk_draw_point (window, half_dark_gc,  x + 1, y + gap_x + gap_width - 1);
      gdk_draw_point (window, dark_gc, x + 1, y + gap_x + gap_width);
    }
    else
    {
      gdk_draw_point (window, half_light_gc, x + 1, y + gap_x + gap_width);
    }
    break;
    
  case GTK_POS_RIGHT:
    gtk_style_apply_default_pixmap (style, window, state_type, area,
				    x + width - 2, y + gap_x, 2, gap_width);
    
    gap_width--; width--; height--;

    gdk_draw_point (window, half_light_gc, x + width, y + gap_x + 1);
    if (gap_x == 0)
    {
      gdk_draw_point (window, light_gc, x + width, y + gap_x);
      gdk_draw_point (window, light_gc, x + width - 1, y + gap_x);
      gdk_draw_point (window, half_light_gc, x + width - 1, y + gap_x + 1);
    }
    else
    {
      gdk_draw_point (window, dark_gc, x + width, y + gap_x);
      gdk_draw_point (window, half_dark_gc,  x + width - 1, y + gap_x);
    }

    gdk_draw_point (window, dark_gc, x + width, y + gap_x + gap_width);
    gdk_draw_point (window, half_dark_gc,  x + width, y + gap_x + gap_width - 1);
    if ((y + gap_x + gap_width) == height)
    {
      gdk_draw_point (window, dark_gc, x + width - 1, y + gap_x + gap_width);
      gdk_draw_point (window, half_dark_gc,  x + width - 1, y + gap_x + gap_width - 1);
    }
    else
      gdk_draw_point (window, half_dark_gc,  x + width - 1, y + gap_x + gap_width);
    break;
  }

  if (area)
  {
    gdk_gc_set_clip_rectangle (light_gc, NULL);
    gdk_gc_set_clip_rectangle (half_light_gc, NULL);
    gdk_gc_set_clip_rectangle (half_dark_gc,  NULL);
    gdk_gc_set_clip_rectangle (dark_gc, NULL);
  }
}  

static void
zenith_draw_extension  (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height,
			GtkPositionType gap_side)
{
  GdkGC           *light_gc, *half_light_gc, *half_dark_gc, *dark_gc;
  ZenithThemeData *zenith_theme_data;
  
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
  
  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  zenith_theme_data = style->engine_data;
  if (zenith_theme_data)
  {
    light_gc      = style->light_gc[state_type];
    half_light_gc = zenith_theme_data->half_light_gc[state_type];
    half_dark_gc  = zenith_theme_data->half_dark_gc[state_type];
    dark_gc       = style->dark_gc[state_type];
  }
  else
  {
    g_warning ("Zenith: zenith_draw_extension: Theme data is NULL");
    light_gc      = style->white_gc;
    half_light_gc = style->light_gc[state_type];
    half_dark_gc  = style->dark_gc[state_type];
    dark_gc       = style->black_gc;
  }

  gtk_paint_box (style, window, state_type, shadow_type, area, widget, detail,
		 x, y, width, height);
  
  if (area)
  {
    gdk_gc_set_clip_rectangle (light_gc, area);
    gdk_gc_set_clip_rectangle (half_light_gc, area);
    gdk_gc_set_clip_rectangle (half_dark_gc,  area);
    gdk_gc_set_clip_rectangle (dark_gc, area);
  }
  
  switch (gap_side) {
  case GTK_POS_TOP:
    gtk_style_apply_default_pixmap (style, window, state_type, area,
				    x, y, width, 2);
    
    width--; height--;
    
    gdk_draw_line_v (window, light_gc, y, y + 1, x);
    gdk_draw_line_v (window, half_light_gc, y, y + 1, x + 1);
    gdk_draw_line_v (window, half_dark_gc,  y, y + 1, x + width - 1);
    gdk_draw_line_v (window, dark_gc, y, y + 1, x + width);
    break;
    
  case GTK_POS_BOTTOM:
    gtk_style_apply_default_pixmap (style, window, state_type, area,
    			    x, y + height - 2, width, 2);
    
    width--; height--;
    
    gdk_draw_line_v (window, light_gc, y + height, y + height - 1, x);
    gdk_draw_line_v (window, half_light_gc, y + height, y + height - 1, x + 1);
    gdk_draw_line_v (window, half_dark_gc,  y + height, y + height - 1, x + width - 1);
    gdk_draw_line_v (window, dark_gc, y + height, y + height - 1, x + width);
    break;
    
  case GTK_POS_LEFT:
    gtk_style_apply_default_pixmap (style, window, state_type, area,
				    x, y, 2, height);
    
    width--; height--;
    
    gdk_draw_line_h (window, light_gc, x, x + 1, y);
    gdk_draw_line_h (window, half_light_gc, x, x + 1, y + 1);
    gdk_draw_line_h (window, half_dark_gc,  x, x + 1, y + height - 1);
    gdk_draw_line_h (window, dark_gc, x, x + 1, y + height);
    break;
    
  case GTK_POS_RIGHT:
    gtk_style_apply_default_pixmap (style, window, state_type, area,
				    x + width - 2, y, 2, height);
    
    width--; height--;
    
    gdk_draw_line_h (window, light_gc, x + width - 2, x + width, y);
    gdk_draw_line_h (window, half_light_gc, x + width - 2, x + width, y + 1);
    gdk_draw_line_h (window, half_dark_gc,  x + width - 2, x + width, y + height - 1);
    gdk_draw_line_h (window, dark_gc, x + width - 2, x + width, y + height);
    break;
  }
  
  if (area)
  {
    gdk_gc_set_clip_rectangle (light_gc, NULL);
    gdk_gc_set_clip_rectangle (half_light_gc, NULL);
    gdk_gc_set_clip_rectangle (half_dark_gc,  NULL);
    gdk_gc_set_clip_rectangle (dark_gc, NULL);
  }
}

static void
zenith_draw_focus      (GtkStyle * style,
			GdkWindow * window,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height)
{
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
}

static void
zenith_draw_slider     (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height,
			GtkOrientation orientation)
{
  g_return_if_fail (style != NULL);
  g_return_if_fail (window != NULL);
  
  if ((width == -1) && (height == -1))
    gdk_window_get_size (window, &width, &height);
  else if (width == -1)
    gdk_window_get_size (window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size (window, NULL, &height);

  gtk_paint_box (style, window, state_type, shadow_type, area, widget, detail,
		 x, y, width, height);

  if (orientation == GTK_ORIENTATION_HORIZONTAL)
    gtk_paint_vline (style, window, state_type, area, widget, detail,
		     y + 2, y + height - 3, width / 2);
  else
    gtk_paint_hline (style, window, state_type, area, widget, detail,
		     x + 2, x + width - 3, height / 2);
}

static void
zenith_draw_handle     (GtkStyle * style,
			GdkWindow * window,
			GtkStateType state_type,
			GtkShadowType shadow_type,
			GdkRectangle * area,
			GtkWidget * widget,
			gchar * detail,
			gint x,
			gint y,
			gint width,
			gint height,
			GtkOrientation orientation)
{
  GdkGC        *light_gc, *dark_gc;
  GdkRectangle  dest;
  gint          xx, yy;
  gint          xthick, ythick;

  g_return_if_fail(style != NULL);
  g_return_if_fail(window != NULL);

  if ((width == -1) && (height == -1))
    gdk_window_get_size(window, &width, &height);
  else if (width == -1)
    gdk_window_get_size(window, &width, NULL);
  else if (height == -1)
    gdk_window_get_size(window, NULL, &height);

  gtk_paint_box (style, window, state_type, shadow_type, area, widget, detail,
		 x, y, width, height);

  light_gc = style->light_gc[state_type];
  dark_gc = style->dark_gc[state_type];

  xthick = style->klass->xthickness;
  ythick = style->klass->ythickness;

  dest.x = x + xthick;
  dest.y = y + ythick;
  dest.width = width - (xthick * 2);
  dest.height = height - (ythick * 2);

  gdk_gc_set_clip_rectangle(light_gc, &dest);
  gdk_gc_set_clip_rectangle(dark_gc,  &dest);

  for (yy = y + ythick; yy < (y + height - ythick); yy += 3)
    for (xx = x + xthick; xx < (x + width - xthick); xx += 6)
    {
      gdk_draw_point (window, light_gc, xx, yy);
      gdk_draw_point (window, dark_gc,  xx + 1, yy + 1);
      
      gdk_draw_point (window, light_gc, xx + 3, yy + 1);
      gdk_draw_point (window, dark_gc,  xx + 4, yy + 2);
    }
  
  gdk_gc_set_clip_rectangle(light_gc, NULL);
  gdk_gc_set_clip_rectangle(dark_gc,  NULL);
}

/*
** Create a Style Class for this theme and load it with the above functions.
*/
GtkStyleClass zenith_class =
{
  2,
  2,
  zenith_draw_hline,
  zenith_draw_vline,
  zenith_draw_shadow,
  zenith_draw_polygon,
  zenith_draw_arrow,
  zenith_draw_diamond,
  zenith_draw_oval,
  zenith_draw_string,
  zenith_draw_box,
  zenith_draw_flat_box,
  zenith_draw_check,
  zenith_draw_option,
  zenith_draw_cross,
  zenith_draw_ramp,
  zenith_draw_tab,
  zenith_draw_shadow_gap,
  zenith_draw_box_gap,
  zenith_draw_extension,
  zenith_draw_focus,
  zenith_draw_slider,
  zenith_draw_handle
};
