/* zenith_theme_main.c - Zenith GTK Theme initialization procedures.
** Copyright (C) 1999 Karl J. Ots.
** 
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
** 
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.
*/

#include <gtk/gtk.h>
#include <gmodule.h>

#include "zenith_theme.h"

#define ZENITH_THEME_DEFAULT_CONTRAST         60
#define ZENITH_THEME_DEFAULT_RADIO_CHECK_SIZE 16
#define ZENITH_THEME_DEFAULT_SCROLLBAR_WIDTH  17
#define ZENITH_THEME_DEFAULT_HSB_FORM         ZENITH_HSB_FORM_NORMAL
#define ZENITH_THEME_DEFAULT_VSB_FORM         ZENITH_VSB_FORM_NORMAL
#define ZENITH_THEME_DEFAULT_SB_PLACE         ZENITH_SB_PLACE_BOTTOM_RIGHT
#define ZENITH_THEME_DEFAULT_SB_INDENT        ZENITH_SB_INDENT_NONE
#define ZENITH_THEME_DEFAULT_HIDE_SB          FALSE

#define SCROLL_DELAY_LENGTH  300
#define SCROLLBAR_SPACING(w) (GTK_SCROLLED_WINDOW_CLASS (GTK_OBJECT (w)->klass)->scrollbar_spacing)
#define RANGE_CLASS(widget)  (GTK_RANGE_CLASS(GTK_OBJECT(widget)->klass))

/*
** Data for the lexical scanner.
*/
enum {
  ZENITH_TOKEN_CONTRAST = G_TOKEN_LAST + 1,
  ZENITH_TOKEN_HSB_FORM,
  ZENITH_TOKEN_VSB_FORM,
  ZENITH_TOKEN_SB_PLACE,
  ZENITH_TOKEN_SB_INDENT,
  ZENITH_TOKEN_SW_FORM,
  ZENITH_TOKEN_HIDE_SB,
  
  ZENITH_TOKEN_LEFT,
  ZENITH_TOKEN_RIGHT,
  ZENITH_TOKEN_TOP,
  ZENITH_TOKEN_BOTTOM,
  ZENITH_TOKEN_TOP_LEFT,
  ZENITH_TOKEN_TOP_RIGHT,
  ZENITH_TOKEN_BOTTOM_LEFT,
  ZENITH_TOKEN_BOTTOM_RIGHT,
  ZENITH_TOKEN_HORIZONTAL,
  ZENITH_TOKEN_VERTICAL,

  ZENITH_TOKEN_NONE,
  ZENITH_TOKEN_TRUE,
  ZENITH_TOKEN_FALSE,
  
  ZENITH_TOKEN_NORMAL,
  ZENITH_TOKEN_PRELIGHT,
  ZENITH_TOKEN_INSENSITIVE,
  ZENITH_TOKEN_ACTIVE,
  ZENITH_TOKEN_SELECTED,
};

/*
** Lookup table for tokens and their names.
*/
static struct { gchar *name; guint token; } zenith_theme_symbols[] = {
  { "contrast",     ZENITH_TOKEN_CONTRAST     },
  { "hsb_form",     ZENITH_TOKEN_HSB_FORM     },
  { "vsb_form",     ZENITH_TOKEN_VSB_FORM     },
  { "sb_placement", ZENITH_TOKEN_SB_PLACE     },
  { "sb_indent",    ZENITH_TOKEN_SB_INDENT    },
  { "sw_form",      ZENITH_TOKEN_SW_FORM      },
  { "hide_sb",      ZENITH_TOKEN_HIDE_SB      },
  
  { "LEFT",         ZENITH_TOKEN_LEFT         },
  { "RIGHT",        ZENITH_TOKEN_RIGHT        },
  { "TOP",          ZENITH_TOKEN_TOP          },
  { "BOTTOM",       ZENITH_TOKEN_BOTTOM       },
  { "TOP_LEFT",     ZENITH_TOKEN_TOP_LEFT     },
  { "TOP_RIGHT",    ZENITH_TOKEN_TOP_RIGHT    },
  { "BOTTOM_LEFT",  ZENITH_TOKEN_BOTTOM_LEFT  },
  { "BOTTOM_RIGHT", ZENITH_TOKEN_BOTTOM_RIGHT },
  { "HORIZONTAL",   ZENITH_TOKEN_HORIZONTAL   },
  { "VERTICAL",     ZENITH_TOKEN_VERTICAL     },

  { "NONE",         ZENITH_TOKEN_NONE         },
  { "TRUE",         ZENITH_TOKEN_TRUE         },
  { "FALSE",        ZENITH_TOKEN_FALSE        },
  
  { "NORMAL",       ZENITH_TOKEN_NORMAL       },
  { "PRELIGHT",     ZENITH_TOKEN_PRELIGHT     },
  { "INSENSITIVE",  ZENITH_TOKEN_INSENSITIVE  },
  { "ACTIVE",       ZENITH_TOKEN_ACTIVE       },
  { "SELECTED",     ZENITH_TOKEN_SELECTED     },
};
static const guint zenith_theme_symbols_num = sizeof (zenith_theme_symbols) /
                                              sizeof (zenith_theme_symbols[0]);

/*
** Functions to export to the outside world
*/
void theme_init (GtkThemeEngine *engine);
void theme_exit (void);

/*
** GtkStyleClass exported from zenith_theme_draw
*/
extern GtkStyleClass zenith_class;

/* The following three functions, rgb_to_hls(), hls_to_rgb() and     */
/* zenith_style_shade() were copied verbatim from the file gtkstyle.c */
/* in the GTK+ source distribution (GTK+ 1.2.1) (zenith_style_shade() */
/* was renamed from gtk_style_shade().                               */

/*
** Copied verbatim from gtkstyle.c.
*/
static void
rgb_to_hls (gdouble *r,
            gdouble *g,
            gdouble *b)
{
  gdouble min;
  gdouble max;
  gdouble red;
  gdouble green;
  gdouble blue;
  gdouble h, l, s;
  gdouble delta;
  
  red = *r;
  green = *g;
  blue = *b;
  
  if (red > green)
    {
      if (red > blue)
        max = red;
      else
        max = blue;
      
      if (green < blue)
        min = green;
      else
        min = blue;
    }
  else
    {
      if (green > blue)
        max = green;
      else
        max = blue;
      
      if (red < blue)
        min = red;
      else
        min = blue;
    }
  
  l = (max + min) / 2;
  s = 0;
  h = 0;
  
  if (max != min)
    {
      if (l <= 0.5)
        s = (max - min) / (max + min);
      else
        s = (max - min) / (2 - max - min);
      
      delta = max -min;
      if (red == max)
        h = (green - blue) / delta;
      else if (green == max)
        h = 2 + (blue - red) / delta;
      else if (blue == max)
        h = 4 + (red - green) / delta;
      
      h *= 60;
      if (h < 0.0)
        h += 360;
    }
  
  *r = h;
  *g = l;
  *b = s;
}

/*
** Copied verbatim from gtkstyle.c.
*/
static void
hls_to_rgb (gdouble *h,
            gdouble *l,
            gdouble *s)
{
  gdouble hue;
  gdouble lightness;
  gdouble saturation;
  gdouble m1, m2;
  gdouble r, g, b;
  
  lightness = *l;
  saturation = *s;
  
  if (lightness <= 0.5)
    m2 = lightness * (1 + saturation);
  else
    m2 = lightness + saturation - lightness * saturation;
  m1 = 2 * lightness - m2;
  
  if (saturation == 0)
    {
      *h = lightness;
      *l = lightness;
      *s = lightness;
    }
  else
    {
      hue = *h + 120;
      while (hue > 360)
        hue -= 360;
      while (hue < 0)
        hue += 360;
      
      if (hue < 60)
        r = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
        r = m2;
      else if (hue < 240)
        r = m1 + (m2 - m1) * (240 - hue) / 60;
      else
        r = m1;
      
      hue = *h;
      while (hue > 360)
        hue -= 360;
      while (hue < 0)
        hue += 360;
      
      if (hue < 60)
        g = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
        g = m2;
      else if (hue < 240)
        g = m1 + (m2 - m1) * (240 - hue) / 60;
      else
        g = m1;
      
      hue = *h - 120;
      while (hue > 360)
        hue -= 360;
      while (hue < 0)
        hue += 360;
      
      if (hue < 60)
        b = m1 + (m2 - m1) * hue / 60;
      else if (hue < 180)
        b = m2;
      else if (hue < 240)
        b = m1 + (m2 - m1) * (240 - hue) / 60;
      else
        b = m1;
      
      *h = r;
      *l = g;
      *s = b;
    }
}

/*
** Copied verbatim from gtkstyle.c (renamed from gtk_style_shade()).
*/
static void
zenith_style_shade (GdkColor *a,
		    GdkColor *b,
		    gdouble   k)
{
  gdouble red;
  gdouble green;
  gdouble blue;
  
  red = (gdouble) a->red / 65535.0;
  green = (gdouble) a->green / 65535.0;
  blue = (gdouble) a->blue / 65535.0;
  
  rgb_to_hls (&red, &green, &blue);
  
  green *= k;
  if (green > 1.0)
    green = 1.0;
  else if (green < 0.0)
    green = 0.0;
  
  blue *= k;
  if (blue > 1.0)
    blue = 1.0;
  else if (blue < 0.0)
    blue = 0.0;
  
  hls_to_rgb (&red, &green, &blue);
  
  b->red = red * 65535.0;
  b->green = green * 65535.0;
  b->blue = blue * 65535.0;
}

/*
** Increase reference count of the given theme data.
*/
static void
zenith_theme_data_ref (ZenithThemeData *zenith_theme_data)
{
  if (zenith_theme_data)
    zenith_theme_data->refcount++;
}

/*
** Create a new ZenithThemeData.  Increase reference count.
*/
static ZenithThemeData *
zenith_theme_data_new (void)
{
  ZenithThemeData *zenith_theme_data = g_new0 (ZenithThemeData, 1);

  zenith_theme_data_ref (zenith_theme_data);
  
  return zenith_theme_data;
}

/*
** Destroy the given theme data.
*/
static void
zenith_theme_data_delete (ZenithThemeData *zenith_theme_data)
{
  g_free (zenith_theme_data);
}

/*
** Decrease the given theme data.  Delete it if the reference
** count falls to zero.
*/
static void
zenith_theme_data_unref (ZenithThemeData *zenith_theme_data)
{
  if (zenith_theme_data)
  {
    zenith_theme_data->refcount--;
    if (zenith_theme_data->refcount == 0)
      zenith_theme_data_delete (zenith_theme_data);
  }
}

/*
** Put the next token from scanner (s) into (t).  If (t) is not token (e),
** then return token (e).
*/
#define return_if_not_next_token(s,t,e) \
{\
  (t) = g_scanner_get_next_token ((s));\
  if ((t) != (e)) return (e);\
}

/*
** Parse contrast statement.
** Syntax:
**   contrast = <integer>
*/
static guint
zenith_theme_parse_contrast (GScanner *scanner,
			     ZenithThemeData *zenith_theme_data)
{
  guint token;

  return_if_not_next_token (scanner, token, ZENITH_TOKEN_CONTRAST);
  return_if_not_next_token (scanner, token, G_TOKEN_EQUAL_SIGN);
  
  return_if_not_next_token (scanner, token, G_TOKEN_INT);
  zenith_theme_data->contrast = scanner->value.v_int;
  
  return G_TOKEN_NONE;
}

/*
** Parse hsb_form statement.
** Syntax:
**   hsb_form = NORMAL|LEFT|RIGHT
*/
static guint
zenith_theme_parse_hsb_form (GScanner *scanner,
			     ZenithThemeData *zenith_theme_data)
{
  guint         token;
  ZenithHSBForm hsb_form;

  return_if_not_next_token (scanner, token, ZENITH_TOKEN_HSB_FORM);
  return_if_not_next_token (scanner, token, G_TOKEN_EQUAL_SIGN);

  token = g_scanner_get_next_token (scanner);
  switch (token) {
  case ZENITH_TOKEN_NORMAL: hsb_form = ZENITH_HSB_FORM_NORMAL; break;
  case ZENITH_TOKEN_LEFT:   hsb_form = ZENITH_HSB_FORM_LEFT;   break;
  case ZENITH_TOKEN_RIGHT:  hsb_form = ZENITH_HSB_FORM_RIGHT;  break;
  default:                  return ZENITH_TOKEN_NORMAL;
  }
  
  zenith_theme_data->hsb_form = hsb_form;
  
  return G_TOKEN_NONE;
}
  
/*
** Parse vsb_form statement.
** Syntax:
**   vsb_form = NORMAL|TOP|BOTTOM
*/
static guint
zenith_theme_parse_vsb_form (GScanner *scanner,
			      ZenithThemeData *zenith_theme_data)
{
  guint         token;
  ZenithVSBForm vsb_form;

  return_if_not_next_token (scanner, token, ZENITH_TOKEN_VSB_FORM);
  return_if_not_next_token (scanner, token, G_TOKEN_EQUAL_SIGN);

  token = g_scanner_get_next_token (scanner);
  switch (token) {
  case ZENITH_TOKEN_NORMAL: vsb_form = ZENITH_VSB_FORM_NORMAL;  break;
  case ZENITH_TOKEN_TOP:    vsb_form = ZENITH_VSB_FORM_TOP;     break;
  case ZENITH_TOKEN_BOTTOM: vsb_form = ZENITH_VSB_FORM_BOTTOM;  break;
  default:                  return ZENITH_TOKEN_NORMAL;
  }
  
  zenith_theme_data->vsb_form = vsb_form;
  
  return G_TOKEN_NONE;
}

/*
** Parse sb_placement statement.
** Syntax:
**   sb_placement = TOPLEFT|TOPRIGHT|BOTTOMLEFT|BOTTOMRIGHT
*/
static guint
zenith_theme_parse_sb_place (GScanner *scanner,
			     ZenithThemeData *zenith_theme_data)
{
  guint         token;
  ZenithSBPlace sb_place;
  
  return_if_not_next_token (scanner, token, ZENITH_TOKEN_SB_PLACE);
  return_if_not_next_token (scanner, token, G_TOKEN_EQUAL_SIGN);
  
  token = g_scanner_get_next_token (scanner);
  switch (token) {
  case ZENITH_TOKEN_TOP_LEFT:     sb_place = ZENITH_SB_PLACE_TOP_LEFT;     break;
  case ZENITH_TOKEN_TOP_RIGHT:    sb_place = ZENITH_SB_PLACE_TOP_RIGHT;    break;
  case ZENITH_TOKEN_BOTTOM_LEFT:  sb_place = ZENITH_SB_PLACE_BOTTOM_LEFT;  break;
  case ZENITH_TOKEN_BOTTOM_RIGHT: sb_place = ZENITH_SB_PLACE_BOTTOM_RIGHT; break;
  default:                        return ZENITH_TOKEN_TOP_LEFT;
  }

  zenith_theme_data->sb_place = sb_place;
  
  return G_TOKEN_NONE;
}

/*
** Parse sb_indent statement.
** Syntax:
**   sb_indent = NONE|HORIZONTAL|VERTICAL
*/
static guint
zenith_theme_parse_sb_indent (GScanner *scanner,
			      ZenithThemeData *zenith_theme_data)
{
  guint          token;
  ZenithSBIndent sb_indent;

  return_if_not_next_token (scanner, token, ZENITH_TOKEN_SB_INDENT);
  return_if_not_next_token (scanner, token, G_TOKEN_EQUAL_SIGN);
  
  token = g_scanner_get_next_token (scanner);
  switch (token) {
  case ZENITH_TOKEN_NONE:       sb_indent = ZENITH_SB_INDENT_NONE;       break;
  case ZENITH_TOKEN_HORIZONTAL: sb_indent = ZENITH_SB_INDENT_HORIZONTAL; break;
  case ZENITH_TOKEN_VERTICAL:   sb_indent = ZENITH_SB_INDENT_VERTICAL;   break;
  default:                      return ZENITH_TOKEN_NONE;
  }
  
  zenith_theme_data->sb_indent = sb_indent;
  
  return G_TOKEN_NONE;
}

/*
** Parse sw_form statement.
** Syntax:
**   sw_form = TOPLEFT|TOPRIGHT|BOTTOMLEFT|BOTTOMRIGHT
*/
static guint
zenith_theme_parse_sw_form (GScanner *scanner,
			    ZenithThemeData *zenith_theme_data)
{
  guint         token;
  ZenithSBPlace sb_place;
  ZenithHSBForm hsb_form;
  ZenithVSBForm vsb_form;
  
  return_if_not_next_token (scanner, token, ZENITH_TOKEN_SW_FORM);
  return_if_not_next_token (scanner, token, G_TOKEN_EQUAL_SIGN);
  
  token = g_scanner_get_next_token (scanner);
  switch (token) {
  case ZENITH_TOKEN_TOP_LEFT:
    sb_place = ZENITH_SB_PLACE_TOP_LEFT;
    hsb_form = ZENITH_HSB_FORM_LEFT;
    vsb_form = ZENITH_VSB_FORM_TOP;
    break;
    
  case ZENITH_TOKEN_TOP_RIGHT:
    sb_place = ZENITH_SB_PLACE_TOP_RIGHT;
    hsb_form = ZENITH_HSB_FORM_RIGHT;
    vsb_form = ZENITH_VSB_FORM_TOP;
    break;
    
  case ZENITH_TOKEN_BOTTOM_LEFT:
    sb_place = ZENITH_SB_PLACE_BOTTOM_LEFT;
    hsb_form = ZENITH_HSB_FORM_LEFT;
    vsb_form = ZENITH_VSB_FORM_BOTTOM;
    break;
    
  case ZENITH_TOKEN_BOTTOM_RIGHT:
    sb_place = ZENITH_SB_PLACE_BOTTOM_RIGHT;
    hsb_form = ZENITH_HSB_FORM_RIGHT;
    vsb_form = ZENITH_VSB_FORM_BOTTOM;
    break;
    
  default:
    return ZENITH_TOKEN_TOP_LEFT;
  }

  zenith_theme_data->sb_place = sb_place;
  zenith_theme_data->hsb_form = hsb_form;
  zenith_theme_data->vsb_form = vsb_form;
  
  return G_TOKEN_NONE;
}

/*
** Parse hide_sb statement.
** Syntax:
**   hide_sb = TRUE|FALSE
*/
static guint
zenith_theme_parse_hide_sb (GScanner *scanner,
			     ZenithThemeData *zenith_theme_data)
{
  guint    token;
  gboolean hide_sb;

  return_if_not_next_token (scanner, token, ZENITH_TOKEN_HIDE_SB);
  return_if_not_next_token (scanner, token, G_TOKEN_EQUAL_SIGN);
  
  token = g_scanner_get_next_token (scanner);
  switch (token) {
  case ZENITH_TOKEN_TRUE:  hide_sb = TRUE;  break;
  case ZENITH_TOKEN_FALSE: hide_sb = FALSE; break;
  default:                 return ZENITH_TOKEN_TRUE;
  }

  zenith_theme_data->hide_sb = hide_sb;
  
  return G_TOKEN_NONE;
}

/*
** Parse engine-specific section of RC file.
*/
static guint
theme_parse_rc_style (GScanner *scanner, GtkRcStyle *rc_style)
{
  static GQuark   scope_id = 0;
  guint           old_scope;
  guint           token;
  ZenithThemeData *zenith_theme_data;
  gint            i;

  /* Setup a new scope in this scanner.  Save the old one for messages */
  if (!scope_id)
    scope_id = g_quark_from_string ("theme_engine");
  old_scope = g_scanner_set_scope (scanner, scope_id);

  /* Add symbols to the symbol table in this scope if it has not */
  /* been done already.                                          */
  if (!g_scanner_lookup_symbol (scanner, zenith_theme_symbols[0].name))
  {
    g_scanner_freeze_symbol_table (scanner);
    for (i = 0; i < zenith_theme_symbols_num; i++)
    {
      g_scanner_scope_add_symbol (scanner, scope_id,
				  zenith_theme_symbols[i].name,
				  GINT_TO_POINTER (zenith_theme_symbols[i].token));
    }
    g_scanner_thaw_symbol_table (scanner);
  }

  /* Let's Go! */
  zenith_theme_data = zenith_theme_data_new ();
  zenith_theme_data->contrast  = ZENITH_THEME_DEFAULT_CONTRAST;
  zenith_theme_data->hsb_form  = ZENITH_THEME_DEFAULT_HSB_FORM;
  zenith_theme_data->vsb_form  = ZENITH_THEME_DEFAULT_VSB_FORM;
  zenith_theme_data->sb_place  = ZENITH_THEME_DEFAULT_SB_PLACE;
  zenith_theme_data->sb_indent = ZENITH_THEME_DEFAULT_SB_INDENT;
  zenith_theme_data->hide_sb   = ZENITH_THEME_DEFAULT_HIDE_SB;

  token = g_scanner_peek_next_token (scanner);
  while (token != G_TOKEN_RIGHT_CURLY)
  {
    switch (token) {
    case ZENITH_TOKEN_CONTRAST:
      token = zenith_theme_parse_contrast (scanner, zenith_theme_data);
      break;
      
    case ZENITH_TOKEN_HSB_FORM:
      token = zenith_theme_parse_hsb_form (scanner, zenith_theme_data);
      break;
      
    case ZENITH_TOKEN_VSB_FORM:
      token = zenith_theme_parse_vsb_form (scanner, zenith_theme_data);
      break;

    case ZENITH_TOKEN_SB_PLACE:
      token = zenith_theme_parse_sb_place (scanner, zenith_theme_data);
      break;
      
    case ZENITH_TOKEN_SB_INDENT:
      token = zenith_theme_parse_sb_indent (scanner, zenith_theme_data);
      break;
      
    case ZENITH_TOKEN_SW_FORM:
      token = zenith_theme_parse_sw_form (scanner, zenith_theme_data);
      break;
      
    case ZENITH_TOKEN_HIDE_SB:
      token = zenith_theme_parse_hide_sb (scanner, zenith_theme_data);
      break;
      
    default:
      g_scanner_get_next_token (scanner);
      token = G_TOKEN_RIGHT_CURLY;
    }

    if (token != G_TOKEN_NONE)
    {
      zenith_theme_data_delete (zenith_theme_data);
      return token;
    }

    token = g_scanner_peek_next_token (scanner);
  }

  g_scanner_get_next_token (scanner);
  rc_style->engine_data = zenith_theme_data;
  g_scanner_set_scope (scanner, old_scope);
  
  return G_TOKEN_NONE;
}

static void
theme_merge_rc_style (GtkRcStyle *dest, GtkRcStyle *src)
{
  ZenithThemeData *zenith_theme_src_data = src->engine_data;

  if (!dest->engine_data && zenith_theme_src_data)
  {
    zenith_theme_data_ref (zenith_theme_src_data);
    dest->engine_data = zenith_theme_src_data;
  }
    
}

static void
theme_rc_style_to_style (GtkStyle *style, GtkRcStyle *rc_style)
{
  ZenithThemeData *zenith_theme_data = rc_style->engine_data;

  if (zenith_theme_data->contrast < 0)
    zenith_theme_data->contrast = 0;
  
  zenith_theme_data_ref (zenith_theme_data);
  style->engine_data = zenith_theme_data;
  
  style->klass = &zenith_class;
}

static void
theme_duplicate_style (GtkStyle *dest, GtkStyle *src)
{
  dest->klass = src->klass;
}

/*
** Adjust the light and dark (and the derived mid) colours.
** (This was mostly copied from step_theme_main.c in GTKStep v1.8).
*/
static void
theme_realize_style (GtkStyle *style)
{
  ZenithThemeData  *zenith_theme_data = style->engine_data;
  GdkGCValuesMask   gc_values_mask = GDK_GC_FOREGROUND;
  GdkGCValues       gc_values;
  gdouble           contrast;
  gint              i;
  
  if (zenith_theme_data)
    contrast = (gdouble) zenith_theme_data->contrast / 100.0;
  else
    contrast = (gdouble) ZENITH_THEME_DEFAULT_CONTRAST / 100.0;

  /* Adjust all the dark/light/mid colours according to the contrast. */
  /* Add half_light/half_dark to theme data                           */
  for (i = 0; i < 5; i++)
  {
    /* Release old GC */
    gtk_gc_release (style->light_gc[i]);
    gtk_gc_release (style->dark_gc[i]);
    gtk_gc_release (style->mid_gc[i]);

    /* Compute new colors */
    zenith_style_shade (&style->bg[i], &style->light[i], 1.0 + contrast);
    zenith_style_shade (&style->bg[i], &style->dark[i],  1.0 - contrast);
    style->mid[i].red   = (style->light[i].red   + style->dark[i].red)   / 2;
    style->mid[i].green = (style->light[i].green + style->dark[i].green) / 2;
    style->mid[i].blue  = (style->light[i].blue  + style->dark[i].blue)  / 2;

    
    /* Allocate new GC */
    if (!gdk_color_alloc (style->colormap, &style->light[i]))
      g_warning ("unable to allocate color: ( %d %d %d )",
		 style->light[i].red, style->light[i].green,
		 style->light[i].blue);
    if (!gdk_color_alloc (style->colormap, &style->dark[i]))
      g_warning ("unable to allocate color: ( %d %d %d )",
		 style->dark[i].red, style->dark[i].green,
		 style->dark[i].blue);
    if (!gdk_color_alloc (style->colormap, &style->mid[i]))
      g_warning ("unable to allocate color: ( %d %d %d )",
		 style->mid[i].red, style->mid[i].green,
		 style->mid[i].blue);
    
    gc_values.foreground = style->light[i];
    style->light_gc[i] = gtk_gc_get (style->depth, style->colormap,
				    &gc_values, gc_values_mask);
    gc_values.foreground = style->dark[i];
    style->dark_gc[i] = gtk_gc_get (style->depth, style->colormap,
				    &gc_values, gc_values_mask);
    
    gc_values.foreground = style->mid[i];
    style->mid_gc[i] = gtk_gc_get (style->depth, style->colormap,
				    &gc_values, gc_values_mask);
    
    if (zenith_theme_data)
    {
      /* Create half-light and half-dark colours */

      zenith_theme_data->half_light[i].red =
	style->bg[i].red + ((style->light[i].red - style->bg[i].red) / 2);
      zenith_theme_data->half_light[i].green =
	style->bg[i].green + ((style->light[i].green - style->bg[i].green) / 2);
      zenith_theme_data->half_light[i].blue =
	style->bg[i].blue + ((style->light[i].blue - style->bg[i].blue) / 2);
      
      zenith_theme_data->half_dark[i].red =
	style->bg[i].red - ((style->bg[i].red - style->dark[i].red) / 2);
      zenith_theme_data->half_dark[i].green =
	style->bg[i].green - ((style->bg[i].green - style->dark[i].green) / 2);
      zenith_theme_data->half_dark[i].blue =
	style->bg[i].blue - ((style->bg[i].blue - style->dark[i].blue) / 2);

    if (!gdk_color_alloc (style->colormap, &zenith_theme_data->half_light[i]))
      g_warning ("unable to allocate color: ( %d %d %d )",
		 zenith_theme_data->half_light[i].red,
		 zenith_theme_data->half_light[i].green,
		 zenith_theme_data->half_light[i].blue);
    if (!gdk_color_alloc (style->colormap, &zenith_theme_data->half_dark[i]))
      g_warning ("unable to allocate color: ( %d %d %d )",
		 zenith_theme_data->half_dark[i].red,
		 zenith_theme_data->half_dark[i].green,
		 zenith_theme_data->half_dark[i].blue);
    
    gc_values.foreground = zenith_theme_data->half_light[i];
    zenith_theme_data->half_light_gc[i] = gtk_gc_get (style->depth, style->colormap,
						      &gc_values, gc_values_mask);
    
    gc_values.foreground = zenith_theme_data->half_dark[i];
    zenith_theme_data->half_dark_gc[i] = gtk_gc_get (style->depth, style->colormap,
						     &gc_values, gc_values_mask);
    }
  }
}

static void
theme_unrealize_style (GtkStyle *style)
{
}

static void
theme_destroy_rc_style (GtkRcStyle *rc_style)
{
  zenith_theme_data_unref (rc_style->engine_data);
}

static void
theme_destroy_style (GtkStyle *style)
{
}

/* Most of the following has been copied directly from the GTK+ */
/* distribution with a few modifications                        */
   
static void
zenith_hscrollbar_calc_slider_size (GtkHScrollbar *hscrollbar)
{
  GtkRange        *range;
  ZenithThemeData *zenith_theme_data;
  ZenithHSBForm    hsb_form;
  gint             step_back_x;
  gint             step_back_width;
  gint             step_forw_x;
  gint             step_forw_width;
  gint             slider_width;
  gint             slider_height;
  gint             left, right;
  gint             width;
  
  g_return_if_fail (hscrollbar != NULL);
  g_return_if_fail (GTK_IS_HSCROLLBAR (hscrollbar));
  
  zenith_theme_data = GTK_WIDGET(hscrollbar)->style->engine_data;
  if (zenith_theme_data)
    hsb_form = zenith_theme_data->hsb_form;
  else
    hsb_form = ZENITH_THEME_DEFAULT_HSB_FORM;
  
  if (GTK_WIDGET_REALIZED (hscrollbar))
  {
    range = GTK_RANGE (hscrollbar);
      
    gdk_window_get_size (range->step_back, &step_back_width, NULL);
    gdk_window_get_size (range->step_forw, &step_forw_width, NULL);
    gdk_window_get_position (range->step_back, &step_back_x, NULL);
    gdk_window_get_position (range->step_forw, &step_forw_x, NULL);

    switch (hsb_form) {
    case ZENITH_HSB_FORM_NORMAL:
      left = (step_back_x +
              step_back_width +
              RANGE_CLASS (hscrollbar)->stepper_slider_spacing);
      right = step_forw_x - RANGE_CLASS (hscrollbar)->stepper_slider_spacing;
      break;
      
    case ZENITH_HSB_FORM_LEFT:
      left = (step_forw_x + RANGE_CLASS (hscrollbar)->stepper_slider_spacing +
	      RANGE_CLASS (hscrollbar)->stepper_size);
      right = (GTK_WIDGET (hscrollbar)->allocation.width -
	       GTK_WIDGET (hscrollbar)->style->klass->xthickness);
      break;
      
    case ZENITH_HSB_FORM_RIGHT:
      left = GTK_WIDGET (hscrollbar)->style->klass->xthickness;
      right = step_back_x - RANGE_CLASS (hscrollbar)->stepper_slider_spacing;
      break;
    }
    width = right - left;
      
      
    if ((range->adjustment->page_size > 0) &&
	(range->adjustment->lower != range->adjustment->upper))
    {
      if (range->adjustment->page_size >
	  (range->adjustment->upper - range->adjustment->lower))
	range->adjustment->page_size = range->adjustment->upper - range->adjustment->lower;
          
      width = (width * range->adjustment->page_size /
	       (range->adjustment->upper - range->adjustment->lower));
          
      if (width < RANGE_CLASS (hscrollbar)->min_slider_size)
	width = RANGE_CLASS (hscrollbar)->min_slider_size;
    }
      
    gdk_window_get_size (range->slider, &slider_width, &slider_height);
      
    if (slider_width != width)
      gdk_window_resize (range->slider, width, slider_height);
  }
}

static void
zenith_vscrollbar_calc_slider_size (GtkVScrollbar *vscrollbar)
{
  GtkRange        *range;
  ZenithThemeData *zenith_theme_data;
  ZenithVSBForm    vsb_form;
  gint             step_back_y;
  gint             step_back_height;
  gint             step_forw_y;
  gint             step_forw_height;
  gint             slider_width;
  gint             slider_height;
  gint             top, bottom;
  gint             height;
  
  g_return_if_fail (vscrollbar != NULL);
  g_return_if_fail (GTK_IS_VSCROLLBAR (vscrollbar));
  
  zenith_theme_data = GTK_WIDGET(vscrollbar)->style->engine_data;
  if (zenith_theme_data)
    vsb_form = zenith_theme_data->vsb_form;
  else
    vsb_form = ZENITH_THEME_DEFAULT_VSB_FORM;
  
  if (GTK_WIDGET_REALIZED (vscrollbar))
  {
    range = GTK_RANGE (vscrollbar);
    
    gdk_window_get_size (range->step_back, NULL, &step_back_height);
    gdk_window_get_size (range->step_forw, NULL, &step_forw_height);
    gdk_window_get_position (range->step_back, NULL, &step_back_y);
    gdk_window_get_position (range->step_forw, NULL, &step_forw_y);

    switch (vsb_form) {
    case ZENITH_VSB_FORM_NORMAL:
      top = (step_back_y + step_back_height +
             RANGE_CLASS (vscrollbar)->stepper_slider_spacing);
      bottom = step_forw_y - RANGE_CLASS (vscrollbar)->stepper_slider_spacing;
      break;
      
    case ZENITH_VSB_FORM_TOP:
      top = (step_forw_y + RANGE_CLASS (vscrollbar)->stepper_slider_spacing +
	     RANGE_CLASS (vscrollbar)->stepper_size);
      bottom = (GTK_WIDGET (vscrollbar)->allocation.height - 
		GTK_WIDGET (vscrollbar)->style->klass->ythickness);
      break;
      
    case ZENITH_VSB_FORM_BOTTOM:
      top = (GTK_WIDGET (vscrollbar)->style->klass->ythickness +
	     RANGE_CLASS (vscrollbar)->stepper_slider_spacing);
      bottom = step_back_y - RANGE_CLASS (vscrollbar)->stepper_slider_spacing;
      break;
    }
    height = bottom - top;
    
    if ((range->adjustment->page_size > 0) &&
	(range->adjustment->lower != range->adjustment->upper))
    {
      if (range->adjustment->page_size >
	  (range->adjustment->upper - range->adjustment->lower))
	range->adjustment->page_size = range->adjustment->upper - range->adjustment->lower;
      
      height = (height * range->adjustment->page_size /
		(range->adjustment->upper - range->adjustment->lower));
      
      if (height < RANGE_CLASS (vscrollbar)->min_slider_size)
	height = RANGE_CLASS (vscrollbar)->min_slider_size;
    }
    
    gdk_window_get_size (range->slider, &slider_width, &slider_height);
    
    if (slider_height != height)
      gdk_window_resize (range->slider, slider_width, height);
  }
}

static void
zenith_range_trough_hdims (GtkRange *range,
			   gint     *left,
			   gint     *right)
{
  ZenithThemeData *zenith_theme_data;
  ZenithHSBForm    hsb_form;
  gint             trough_width;
  gint             slider_length;
  gint             tmp_width;
  gint             tleft;
  gint             tright;

  g_return_if_fail (range != NULL);

  zenith_theme_data = GTK_WIDGET(range)->style->engine_data;
  if (zenith_theme_data)
    hsb_form = zenith_theme_data->hsb_form;
  else
    hsb_form = ZENITH_THEME_DEFAULT_HSB_FORM;
  
  gdk_window_get_size (range->trough, &trough_width, NULL);
  gdk_window_get_size (range->slider, &slider_length, NULL);

  tleft = GTK_WIDGET (range)->style->klass->xthickness;
  tright = trough_width - slider_length - GTK_WIDGET (range)->style->klass->xthickness;

  if (range->step_back)
  {
    gdk_window_get_size (range->step_back, &tmp_width, NULL);
    switch (hsb_form) {
    case ZENITH_HSB_FORM_NORMAL:
      tleft += (tmp_width + RANGE_CLASS (range)->stepper_slider_spacing);
      break;
      
    case ZENITH_HSB_FORM_LEFT:
      tleft += tmp_width;
      break;
      
    case ZENITH_HSB_FORM_RIGHT:
      tright -= (tmp_width + RANGE_CLASS (range)->stepper_slider_spacing);
      break;
    }
  }
  
  if (range->step_forw)
  {
    gdk_window_get_size (range->step_forw, &tmp_width, NULL);
    switch (hsb_form) {
    case ZENITH_HSB_FORM_NORMAL:
      tright -= (tmp_width + RANGE_CLASS (range)->stepper_slider_spacing);
      break;
      
    case ZENITH_HSB_FORM_LEFT:
      tleft += (tmp_width + RANGE_CLASS (range)->stepper_slider_spacing);
      break;
      
    case ZENITH_HSB_FORM_RIGHT:
      tright -= tmp_width;
      break;
    }
  }
  
  if (left)
    *left = tleft;
  if (right)
    *right = tright;
}

static void
zenith_range_trough_vdims (GtkRange *range,
			   gint     *top,
			   gint     *bottom)
{
  ZenithThemeData *zenith_theme_data;
  ZenithVSBForm    vsb_form;
  gint             trough_height;
  gint             slider_length;
  gint             tmp_height;
  gint             ttop;
  gint             tbottom;

  g_return_if_fail (range != NULL);

  zenith_theme_data = GTK_WIDGET(range)->style->engine_data;
  if (zenith_theme_data)
    vsb_form = zenith_theme_data->vsb_form;
  else
    vsb_form = ZENITH_THEME_DEFAULT_VSB_FORM;
  
  gdk_window_get_size (range->trough, NULL, &trough_height);
  gdk_window_get_size (range->slider, NULL, &slider_length);

  ttop = GTK_WIDGET (range)->style->klass->ythickness;
  tbottom = trough_height - slider_length - GTK_WIDGET (range)->style->klass->ythickness;

  if (range->step_back)
  {
    gdk_window_get_size (range->step_back, NULL, &tmp_height);
    switch (vsb_form) {
    case ZENITH_VSB_FORM_NORMAL:
      ttop += (tmp_height + RANGE_CLASS (range)->stepper_slider_spacing);
      break;
      
    case ZENITH_VSB_FORM_TOP:
      ttop += tmp_height;
      break;
      
    case ZENITH_VSB_FORM_BOTTOM:
      tbottom -= (tmp_height + RANGE_CLASS (range)->stepper_slider_spacing);
      break;
    }
  }
  
  if (range->step_forw)
  {
    gdk_window_get_size (range->step_forw, NULL, &tmp_height);
    switch (vsb_form) {
    case ZENITH_VSB_FORM_NORMAL:
      tbottom -= (tmp_height + RANGE_CLASS (range)->stepper_slider_spacing);
      break;
      
    case ZENITH_VSB_FORM_TOP:
      ttop += (tmp_height + RANGE_CLASS (range)->stepper_slider_spacing);
      break;
      
    case ZENITH_VSB_FORM_BOTTOM:
      tbottom -= tmp_height;
      break;
    }
  }

  if (top)
    *top = ttop;
  if (bottom)
    *bottom = tbottom;
}

static void
zenith_range_remove_timer (GtkRange *range)
{
  g_return_if_fail (range != NULL);
  g_return_if_fail (GTK_IS_RANGE (range));

  if (range->timer)
  {
    gtk_timeout_remove (range->timer);
    range->timer = 0;
  }
  range->need_timer = FALSE;
}

static gint
zenith_range_htrough_click (GtkRange *range,
			    gint      x,
			    gint      y,
			    gfloat   *jump_perc)
{
  gint ythickness;
  gint trough_width;
  gint trough_height;
  gint slider_x;
  gint slider_length;
  gint left, right;
  
  g_return_val_if_fail (range != NULL, GTK_TROUGH_NONE);
  g_return_val_if_fail (GTK_IS_RANGE (range), GTK_TROUGH_NONE);
  
  ythickness = GTK_WIDGET (range)->style->klass->ythickness;

  zenith_range_trough_hdims (range, &left, &right);
  gdk_window_get_size (range->slider, &slider_length, NULL);
  right += slider_length;
  
  if ((x > left) && (y > ythickness))
  {
    gdk_window_get_size (range->trough, &trough_width, &trough_height);
    
    if ((x < right) && (y < (trough_height - ythickness)))
    {
      if (jump_perc)
      {
	*jump_perc = ((gdouble) (x - left)) / ((gdouble) (right - left));
	return GTK_TROUGH_JUMP;
      }
      
      gdk_window_get_position (range->slider, &slider_x, NULL);
      
      if (x < slider_x)
	return GTK_TROUGH_START;
      else
	return GTK_TROUGH_END;
    }
  }
  
  return GTK_TROUGH_NONE;
}

static gint
zenith_range_vtrough_click (GtkRange *range,
			    gint      x,
			    gint      y,
			    gfloat   *jump_perc)
{
  gint xthickness;
  gint trough_width;
  gint trough_height;
  gint slider_y;
  gint top, bottom;
  gint slider_length;
  
  g_return_val_if_fail (range != NULL, GTK_TROUGH_NONE);
  g_return_val_if_fail (GTK_IS_RANGE (range), GTK_TROUGH_NONE);
  
  xthickness = GTK_WIDGET (range)->style->klass->xthickness;
  
  zenith_range_trough_vdims (range, &top, &bottom);
  gdk_window_get_size (range->slider, NULL, &slider_length);
  bottom += slider_length;
  
  if ((x > xthickness) && (y > top))
  {
    gdk_window_get_size (range->trough, &trough_width, &trough_height);
    
    if ((x < (trough_width - xthickness) && (y < bottom)))
    {
      if (jump_perc)
      {
	*jump_perc = ((gdouble) (y - top)) / ((gdouble) (bottom - top));
	
	return GTK_TROUGH_JUMP;
      }
      
      gdk_window_get_position (range->slider, NULL, &slider_y);
      
      if (y < slider_y)
	return GTK_TROUGH_START;
      else
	return GTK_TROUGH_END;
    }
  }
  
  return GTK_TROUGH_NONE;
}

static void
zenith_range_hmotion (GtkRange *range,
		      gint      xdelta,
		      gint      ydelta)
{
  gdouble old_value;
  gint left, right;
  gint slider_x, slider_y;
  gint new_pos;

  g_return_if_fail (range != NULL);
  g_return_if_fail (GTK_IS_RANGE (range));
  
  range = GTK_RANGE (range);
  
  gdk_window_get_position (range->slider, &slider_x, &slider_y);
  zenith_range_trough_hdims (range, &left, &right);

  if (left == right)
    return;

  new_pos = slider_x + xdelta;

  if (new_pos < left)
    new_pos = left;
  else if (new_pos > right)
    new_pos = right;

  old_value = range->adjustment->value;
  range->adjustment->value = ((range->adjustment->upper -
			       range->adjustment->lower -
			       range->adjustment->page_size) *
			      (new_pos - left) / (right - left) +
			      range->adjustment->lower);

  if (range->digits >= 0)
  {
    char buffer[64];
    
    sprintf (buffer, "%0.*f", range->digits, range->adjustment->value);
    sscanf (buffer, "%f", &range->adjustment->value);
  }
  
  if (old_value != range->adjustment->value)
  {
    if (range->policy == GTK_UPDATE_CONTINUOUS)
    {
      gtk_signal_emit_by_name (GTK_OBJECT (range->adjustment), "value_changed");
    }
    else
    {
      gtk_range_slider_update (range);
      gtk_range_clear_background (range);
      
      if (range->policy == GTK_UPDATE_DELAYED)
      {
	zenith_range_remove_timer (range);
	range->timer = gtk_timeout_add (SCROLL_DELAY_LENGTH,
					(GtkFunction) RANGE_CLASS (range)->timer,
					(gpointer) range);
      }
    }
  }
}

static void
zenith_range_vmotion (GtkRange *range,
		      gint      xdelta,
		      gint      ydelta)
{
  gdouble old_value;
  gint top, bottom;
  gint slider_x, slider_y;
  gint new_pos;
  
  g_return_if_fail (range != NULL);
  g_return_if_fail (GTK_IS_RANGE (range));
  
  range = GTK_RANGE (range);
  
  gdk_window_get_position (range->slider, &slider_x, &slider_y);
  zenith_range_trough_vdims (range, &top, &bottom);

  if (bottom == top)
    return;

  new_pos = slider_y + ydelta;

  if (new_pos < top)
    new_pos = top;
  else if (new_pos > bottom)
    new_pos = bottom;
  
  old_value = range->adjustment->value;
  range->adjustment->value = ((range->adjustment->upper -
			       range->adjustment->lower -
			       range->adjustment->page_size) *
			      (new_pos - top) / (bottom - top) +
			      range->adjustment->lower);
  
  if (range->digits >= 0)
  {
    char buffer[64];
    
    sprintf (buffer, "%0.*f", range->digits, range->adjustment->value);
    sscanf (buffer, "%f", &range->adjustment->value);
  }
  
  if (old_value != range->adjustment->value)
  {
    if (range->policy == GTK_UPDATE_CONTINUOUS)
    {
      gtk_signal_emit_by_name (GTK_OBJECT (range->adjustment), "value_changed");
    }
    else
    {
      gtk_range_slider_update (range);
      gtk_range_clear_background (range);
      
      if (range->policy == GTK_UPDATE_DELAYED)
      {
	zenith_range_remove_timer (range);
	range->timer = gtk_timeout_add (SCROLL_DELAY_LENGTH,
					(GtkFunction) RANGE_CLASS (range)->timer,
					(gpointer) range);
      }
    }
  }
}

static void
zenith_range_show_hide (GtkRange *range)
{
  ZenithThemeData *zenith_theme_data;
  gboolean hide_sb;
  
  g_return_if_fail (range != NULL);
  g_return_if_fail (GTK_IS_RANGE (range));

  zenith_theme_data = GTK_WIDGET (range)->style->engine_data;
  if (zenith_theme_data)
    hide_sb = zenith_theme_data->hide_sb;
  else
    hide_sb = ZENITH_THEME_DEFAULT_HIDE_SB;

  if (hide_sb && (range->adjustment->page_size ==
		  (range->adjustment->upper - range->adjustment->lower)))
  {
    gdk_window_hide (range->slider);
    gdk_window_hide (range->step_back);
    gdk_window_hide (range->step_forw);
  }
  else
  {
    gdk_window_show (range->slider);
    gdk_window_show (range->step_back);
    gdk_window_show (range->step_forw);
  }
}
    
static void
zenith_range_hslider_update (GtkRange *range)
{
  gint left;
  gint right;
  gint x;

  g_return_if_fail (range != NULL);
  g_return_if_fail (GTK_IS_RANGE (range));

  if (GTK_WIDGET_REALIZED (range))
  {
    zenith_range_trough_hdims (range, &left, &right);
    x = left;
    
    if (range->adjustment->value < range->adjustment->lower)
    {
      range->adjustment->value = range->adjustment->lower;
      gtk_signal_emit_by_name (GTK_OBJECT (range->adjustment), "value_changed");
    }
    else if (range->adjustment->value > range->adjustment->upper)
    {
      range->adjustment->value = range->adjustment->upper;
      gtk_signal_emit_by_name (GTK_OBJECT (range->adjustment), "value_changed");
    }
    
    if (range->adjustment->lower != (range->adjustment->upper - range->adjustment->page_size))
      x += ((right - left) * (range->adjustment->value - range->adjustment->lower) /
	    (range->adjustment->upper - range->adjustment->lower - range->adjustment->page_size));
    
    if (x < left)
      x = left;
    else if (x > right)
      x = right;
    
    zenith_range_show_hide (range);
      
    gdk_window_move (range->slider, x, GTK_WIDGET (range)->style->klass->ythickness);
  }
}

void
zenith_range_vslider_update (GtkRange *range)
{
  gint top;
  gint bottom;
  gint y;
  
  g_return_if_fail (range != NULL);
  g_return_if_fail (GTK_IS_RANGE (range));
  
  if (GTK_WIDGET_REALIZED (range))
  {
    zenith_range_trough_vdims (range, &top, &bottom);
    y = top;
    
    if (range->adjustment->value < range->adjustment->lower)
    {
      range->adjustment->value = range->adjustment->lower;
      gtk_signal_emit_by_name (GTK_OBJECT (range->adjustment), "value_changed");
    }
    else if (range->adjustment->value > range->adjustment->upper)
    {
      range->adjustment->value = range->adjustment->upper;
      gtk_signal_emit_by_name (GTK_OBJECT (range->adjustment), "value_changed");
    }
    
    if (range->adjustment->lower != (range->adjustment->upper - range->adjustment->page_size))
      y += ((bottom - top) * (range->adjustment->value - range->adjustment->lower) /
	    (range->adjustment->upper - range->adjustment->lower - range->adjustment->page_size));
    
    if (y < top)
      y = top;
    else if (y > bottom)
      y = bottom;
    
    zenith_range_show_hide (range);
      
    gdk_window_move (range->slider, GTK_WIDGET (range)->style->klass->xthickness, y);
  }
}

static void
zenith_hscrollbar_slider_update (GtkRange *range)
{
  g_return_if_fail (range != NULL);
  g_return_if_fail (GTK_IS_HSCROLLBAR (range));

  zenith_hscrollbar_calc_slider_size (GTK_HSCROLLBAR (range));
  zenith_range_hslider_update (range);
}

static void
zenith_vscrollbar_slider_update (GtkRange *range)
{
  g_return_if_fail (range != NULL);
  g_return_if_fail (GTK_IS_VSCROLLBAR (range));
  
  zenith_vscrollbar_calc_slider_size (GTK_VSCROLLBAR (range));
  zenith_range_vslider_update (range);
}

static void
zenith_hscrollbar_realize (GtkWidget *widget)
{
  GtkRange        *range;
  GdkWindowAttr    attributes;
  ZenithThemeData *zenith_theme_data;
  ZenithHSBForm    hsb_form;
  gint             attributes_mask;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_HSCROLLBAR (widget));
  
  zenith_theme_data = widget->style->engine_data;
  if (zenith_theme_data)
    hsb_form = zenith_theme_data->hsb_form;
  else
    hsb_form = ZENITH_THEME_DEFAULT_HSB_FORM;
  
  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  range = GTK_RANGE (widget);
  
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y + (widget->allocation.height - widget->requisition.height) / 2;
  attributes.width = widget->allocation.width;
  attributes.height = widget->requisition.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
                            GDK_BUTTON_PRESS_MASK |
                            GDK_BUTTON_RELEASE_MASK |
                            GDK_ENTER_NOTIFY_MASK |
                            GDK_LEAVE_NOTIFY_MASK);
  
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
  
  range->trough = widget->window;
  gdk_window_ref (range->trough);
  
  switch (hsb_form) {
  case ZENITH_HSB_FORM_NORMAL:
    attributes.x = widget->style->klass->xthickness;
    break;
    
  case ZENITH_HSB_FORM_LEFT:
    attributes.x = widget->style->klass->xthickness;
    break;
    
  case ZENITH_HSB_FORM_RIGHT:
    attributes.x = (widget->allocation.width -
		    widget->style->klass->xthickness -
		    2 * RANGE_CLASS (widget)->stepper_size);
    break;
  }
  
  attributes.y = widget->style->klass->ythickness;
  attributes.width = RANGE_CLASS (widget)->stepper_size;
  attributes.height = RANGE_CLASS (widget)->stepper_size;
  
  range->step_back = gdk_window_new (range->trough, &attributes, attributes_mask);
  
  switch (hsb_form) {
  case ZENITH_HSB_FORM_NORMAL:
    attributes.x = (widget->allocation.width -
		    widget->style->klass->xthickness -
		    RANGE_CLASS (widget)->stepper_size);
    break;
    
  case ZENITH_HSB_FORM_LEFT:
    attributes.x = (widget->style->klass->xthickness +
		    RANGE_CLASS (widget)->stepper_size);
    break;
    
  case ZENITH_HSB_FORM_RIGHT:
    attributes.x = (widget->allocation.width -
		    widget->style->klass->xthickness -
		    RANGE_CLASS (widget)->stepper_size);
    break;
  }
  
  range->step_forw = gdk_window_new (range->trough, &attributes, attributes_mask);
  
  attributes.x = 0;
  attributes.y = widget->style->klass->ythickness;
  attributes.width = RANGE_CLASS (widget)->min_slider_size;
  attributes.height = RANGE_CLASS (widget)->slider_width;
  attributes.event_mask |= (GDK_BUTTON_MOTION_MASK |
                            GDK_POINTER_MOTION_HINT_MASK);
  
  range->slider = gdk_window_new (range->trough, &attributes, attributes_mask);
  
  zenith_hscrollbar_calc_slider_size (GTK_HSCROLLBAR (widget));
  gtk_range_slider_update (GTK_RANGE (widget));
  
  widget->style = gtk_style_attach (widget->style, widget->window);
  
  gdk_window_set_user_data (range->trough, widget);
  gdk_window_set_user_data (range->slider, widget);
  gdk_window_set_user_data (range->step_forw, widget);
  gdk_window_set_user_data (range->step_back, widget);
  
  gtk_style_set_background (widget->style, range->trough, GTK_STATE_ACTIVE);
  gtk_style_set_background (widget->style, range->slider, GTK_STATE_NORMAL);
  gtk_style_set_background (widget->style, range->step_forw, GTK_STATE_ACTIVE);
  gtk_style_set_background (widget->style, range->step_back, GTK_STATE_ACTIVE);
  
  gdk_window_show (range->slider);
  gdk_window_show (range->step_forw);
  gdk_window_show (range->step_back);
  
  zenith_range_show_hide (range);
}

static void
zenith_vscrollbar_realize (GtkWidget *widget)
{
  GtkRange        *range;
  GdkWindowAttr    attributes;
  ZenithThemeData *zenith_theme_data;
  ZenithVSBForm    vsb_form;
  gint             attributes_mask;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_VSCROLLBAR (widget));
  
  zenith_theme_data = widget->style->engine_data;
  if (zenith_theme_data)
    vsb_form = zenith_theme_data->vsb_form;
  else
    vsb_form = ZENITH_THEME_DEFAULT_VSB_FORM;
  
  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  range = GTK_RANGE (widget);
  
  attributes.x = widget->allocation.x + (widget->allocation.width - widget->requisition.width) / 2;
  attributes.y = widget->allocation.y;
  attributes.width = widget->requisition.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
                            GDK_BUTTON_PRESS_MASK |
                            GDK_BUTTON_RELEASE_MASK |
                            GDK_ENTER_NOTIFY_MASK |
                            GDK_LEAVE_NOTIFY_MASK);
  
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
  
  range->trough = widget->window;
  gdk_window_ref (range->trough);
  
  attributes.x = widget->style->klass->xthickness;
  switch (vsb_form) {
  case ZENITH_VSB_FORM_NORMAL:
    attributes.y = widget->style->klass->ythickness;
    break;
      
  case ZENITH_VSB_FORM_TOP:
    attributes.y = widget->style->klass->ythickness;
    break;
      
  case ZENITH_VSB_FORM_BOTTOM:
    attributes.y = (widget->allocation.height -
		    widget->style->klass->ythickness -
		    2 * RANGE_CLASS (widget)->stepper_size);
    break;
  }
  attributes.width = RANGE_CLASS (widget)->stepper_size;
  attributes.height = RANGE_CLASS (widget)->stepper_size;
  
  range->step_back = gdk_window_new (range->trough, &attributes, attributes_mask);
  
  switch (vsb_form) {
  case ZENITH_VSB_FORM_NORMAL:
    attributes.y = (widget->allocation.height -
		    widget->style->klass->ythickness -
		    RANGE_CLASS (widget)->stepper_size);
    break;
    
  case ZENITH_VSB_FORM_TOP:
    attributes.y = (widget->style->klass->ythickness +
		    RANGE_CLASS (widget)->stepper_size);
    break;
      
  case ZENITH_VSB_FORM_BOTTOM:
    attributes.y = (widget->allocation.height -
		    widget->style->klass->ythickness -
		    RANGE_CLASS (widget)->stepper_size);
    break;
  }
  
  range->step_forw = gdk_window_new (range->trough, &attributes, attributes_mask);
  
  attributes.x = widget->style->klass->ythickness;
  attributes.y = 0;
  attributes.width = RANGE_CLASS (widget)->slider_width;
  attributes.height = RANGE_CLASS (widget)->min_slider_size;
  attributes.event_mask |= (GDK_BUTTON_MOTION_MASK |
                            GDK_POINTER_MOTION_HINT_MASK);
  
  range->slider = gdk_window_new (range->trough, &attributes, attributes_mask);
  
  zenith_vscrollbar_calc_slider_size (GTK_VSCROLLBAR (widget));
  gtk_range_slider_update (GTK_RANGE (widget));
  
  widget->style = gtk_style_attach (widget->style, widget->window);
  
  gdk_window_set_user_data (range->trough, widget);
  gdk_window_set_user_data (range->slider, widget);
  gdk_window_set_user_data (range->step_forw, widget);
  gdk_window_set_user_data (range->step_back, widget);
  
  gtk_style_set_background (widget->style, range->trough, GTK_STATE_ACTIVE);
  gtk_style_set_background (widget->style, range->slider, GTK_STATE_NORMAL);
  gtk_style_set_background (widget->style, range->step_forw, GTK_STATE_ACTIVE);
  gtk_style_set_background (widget->style, range->step_back, GTK_STATE_ACTIVE);
  
  gdk_window_show (range->slider);
  gdk_window_show (range->step_forw);
  gdk_window_show (range->step_back);
  
  zenith_range_show_hide (range);
}

static void
zenith_hscrollbar_size_allocate (GtkWidget     *widget,
				 GtkAllocation *allocation)
{
  GtkRange        *range;
  ZenithThemeData *zenith_theme_data;
  ZenithHSBForm    hsb_form;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_HSCROLLBAR (widget));
  g_return_if_fail (allocation != NULL);

  zenith_theme_data = widget->style->engine_data;
  if (zenith_theme_data)
    hsb_form = zenith_theme_data->hsb_form;
  else
    hsb_form = ZENITH_THEME_DEFAULT_HSB_FORM;
  
  widget->allocation = *allocation;
  if (GTK_WIDGET_REALIZED (widget))
  {
    range = GTK_RANGE (widget);
    
    gdk_window_move_resize (range->trough,
			    allocation->x,
			    allocation->y + (allocation->height - widget->requisition.height) / 2,
			    allocation->width, widget->requisition.height);
    
    switch (hsb_form) {
    case ZENITH_HSB_FORM_NORMAL:
      gdk_window_move_resize (range->step_back,
                              widget->style->klass->xthickness,
                              widget->style->klass->ythickness,
			      RANGE_CLASS (widget)->stepper_size,
                              widget->requisition.height - widget->style->klass->ythickness * 2);
      gdk_window_move_resize (range->step_forw,
                              allocation->width - widget->style->klass->xthickness - RANGE_CLASS (widget)->stepper_size,
                              widget->style->klass->ythickness,
                              RANGE_CLASS (widget)->stepper_size,
                              widget->requisition.height - widget->style->klass->ythickness * 2);
      break;
      
    case ZENITH_HSB_FORM_LEFT:
      break;
      gdk_window_move_resize (range->step_back,
                              widget->style->klass->xthickness,
                              widget->style->klass->ythickness,
                              RANGE_CLASS (widget)->stepper_size,
                              widget->requisition.height - widget->style->klass->ythickness * 2);
      gdk_window_move_resize (range->step_forw,
                              widget->style->klass->xthickness + RANGE_CLASS (widget)->stepper_size,
                              widget->style->klass->ythickness,
                              RANGE_CLASS (widget)->stepper_size,
                              widget->requisition.height - widget->style->klass->ythickness * 2);
    
    case ZENITH_HSB_FORM_RIGHT:
      gdk_window_move_resize (range->step_back,
			      allocation->width - widget->style->klass->xthickness - 2 * RANGE_CLASS (widget)->stepper_size,
			      widget->style->klass->ythickness,
			      RANGE_CLASS (widget)->stepper_size,
			      widget->requisition.height - widget->style->klass->ythickness * 2);
      gdk_window_move_resize (range->step_forw,
			      allocation->width - widget->style->klass->xthickness - RANGE_CLASS (widget)->stepper_size,
			      widget->style->klass->ythickness,
			      RANGE_CLASS (widget)->stepper_size,
			      widget->requisition.height - widget->style->klass->ythickness * 2);
      break;
    }
    
    gdk_window_resize (range->slider,
		       RANGE_CLASS (widget)->min_slider_size,
		       widget->requisition.height - widget->style->klass->ythickness * 2);
    
    gtk_range_slider_update (GTK_RANGE (widget));
  }
}

static void
zenith_vscrollbar_size_allocate (GtkWidget     *widget,
                              GtkAllocation *allocation)
{
  GtkRange        *range;
  ZenithThemeData *zenith_theme_data;
  ZenithVSBForm    vsb_form;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_VSCROLLBAR (widget));
  g_return_if_fail (allocation != NULL);
  
  zenith_theme_data = widget->style->engine_data;
  if (zenith_theme_data)
    vsb_form = zenith_theme_data->vsb_form;
  else
    vsb_form = ZENITH_THEME_DEFAULT_VSB_FORM;
  
  widget->allocation = *allocation;
  if (GTK_WIDGET_REALIZED (widget))
  {
    range = GTK_RANGE (widget);
    
    gdk_window_move_resize (range->trough,
			    allocation->x + (allocation->width - widget->requisition.width) / 2,
			    allocation->y,
			    widget->requisition.width, allocation->height);
    switch (vsb_form) {
    case ZENITH_VSB_FORM_NORMAL:
      gdk_window_move_resize (range->step_back,
                              widget->style->klass->xthickness,
                              widget->style->klass->ythickness,
                              widget->requisition.width - widget->style->klass->xthickness * 2,
                              RANGE_CLASS (widget)->stepper_size);
      gdk_window_move_resize (range->step_forw,
                              widget->style->klass->xthickness,
                              allocation->height - widget->style->klass->ythickness - RANGE_CLASS (widget)->stepper_size,
                              widget->requisition.width - widget->style->klass->xthickness * 2,
                              RANGE_CLASS (widget)->stepper_size);
      break;
      
    case ZENITH_VSB_FORM_TOP:
      gdk_window_move_resize (range->step_back,
                              widget->style->klass->xthickness,
                              widget->style->klass->ythickness,
                              widget->requisition.width - widget->style->klass->xthickness * 2,
                              RANGE_CLASS (widget)->stepper_size);
      gdk_window_move_resize (range->step_forw,
                              widget->style->klass->xthickness,
                              widget->style->klass->ythickness + RANGE_CLASS (widget)->stepper_size,
                              widget->requisition.width - widget->style->klass->xthickness * 2,
                              RANGE_CLASS (widget)->stepper_size);
      break;
      
    case ZENITH_VSB_FORM_BOTTOM:
      gdk_window_move_resize (range->step_back,
			      widget->style->klass->xthickness,
			      allocation->height - widget->style->klass->ythickness - 2 * RANGE_CLASS (widget)->stepper_size,
			      widget->requisition.width - widget->style->klass->xthickness * 2,
			      RANGE_CLASS (widget)->stepper_size);
      gdk_window_move_resize (range->step_forw,
			      widget->style->klass->xthickness,
			      allocation->height - widget->style->klass->ythickness - RANGE_CLASS (widget)->stepper_size,
			      widget->requisition.width - widget->style->klass->xthickness * 2,
			      RANGE_CLASS (widget)->stepper_size);
      break;
    }
    
    gdk_window_resize (range->slider,
		       widget->requisition.width - widget->style->klass->xthickness * 2,
		       RANGE_CLASS (range)->min_slider_size);
    
    gtk_range_slider_update (GTK_RANGE (widget));
  }
}

static void
zenith_scrolled_window_set_placement (GtkWidget *widget)
{
  ZenithThemeData *zenith_theme_data;
  ZenithSBPlace    sb_place;
  GtkCornerType    corner;

  if (widget && GTK_IS_SCROLLED_WINDOW(widget))
  {
    zenith_theme_data = widget->style->engine_data;
    if (zenith_theme_data)
      sb_place = zenith_theme_data->sb_place;
    else
    {
      g_warning ("zenith_scrolled_window_set_placement: Theme Data is NULL");
      sb_place = ZENITH_THEME_DEFAULT_SB_PLACE;
    }

    switch (sb_place) {
    case ZENITH_SB_PLACE_TOP_LEFT:
      corner = GTK_CORNER_BOTTOM_RIGHT;
      break;
      
    case ZENITH_SB_PLACE_BOTTOM_LEFT:
      corner = GTK_CORNER_TOP_RIGHT;
      break;
      
    case ZENITH_SB_PLACE_TOP_RIGHT:
      corner = GTK_CORNER_BOTTOM_LEFT;
      break;
      
    case ZENITH_SB_PLACE_BOTTOM_RIGHT:
      corner = GTK_CORNER_TOP_LEFT;
      break;
    }
    
    gtk_scrolled_window_set_placement (GTK_SCROLLED_WINDOW (widget), corner);
  }
}

static void
zenith_scrolled_window_relative_allocation (GtkWidget     *widget,
					    GtkAllocation *allocation)
{
  GtkScrolledWindow *scrolled_window;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (allocation != NULL);
  
  scrolled_window = GTK_SCROLLED_WINDOW (widget);
  
  allocation->x = GTK_CONTAINER (widget)->border_width;
  allocation->y = GTK_CONTAINER (widget)->border_width;
  allocation->width = MAX (1, (gint)widget->allocation.width - allocation->x * 2);
  allocation->height = MAX (1, (gint)widget->allocation.height - allocation->y * 2);
  
  if (scrolled_window->vscrollbar_visible)
  {
    GtkRequisition vscrollbar_requisition;
    gtk_widget_get_child_requisition (scrolled_window->vscrollbar,
				      &vscrollbar_requisition);
    
    if (scrolled_window->window_placement == GTK_CORNER_TOP_RIGHT ||
	scrolled_window->window_placement == GTK_CORNER_BOTTOM_RIGHT)
      allocation->x += (vscrollbar_requisition.width +
			SCROLLBAR_SPACING (scrolled_window));
    
    allocation->width = MAX (1, (gint)allocation->width -
			     ((gint)vscrollbar_requisition.width +
			      (gint)SCROLLBAR_SPACING (scrolled_window)));
  }
  if (scrolled_window->hscrollbar_visible)
  {
    GtkRequisition hscrollbar_requisition;
    gtk_widget_get_child_requisition (scrolled_window->hscrollbar,
				      &hscrollbar_requisition);
    
    if (scrolled_window->window_placement == GTK_CORNER_BOTTOM_LEFT ||
	scrolled_window->window_placement == GTK_CORNER_BOTTOM_RIGHT)
      allocation->y += (hscrollbar_requisition.height +
			SCROLLBAR_SPACING (scrolled_window));
    
    allocation->height = MAX (1, (gint)allocation->height -
			      ((gint)hscrollbar_requisition.height +
			       (gint)SCROLLBAR_SPACING (scrolled_window)));
  }
}

static void
zenith_scrolled_window_size_allocate (GtkWidget     *widget,
				      GtkAllocation *allocation)
{
  GtkScrolledWindow *scrolled_window;
  GtkBin            *bin;
  GtkAllocation     relative_allocation;
  GtkAllocation     child_allocation;
  ZenithThemeData  *zenith_theme_data;
  ZenithSBPlace     sb_place;
  ZenithSBIndent    sb_indent;
  GtkCornerType     corner;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_SCROLLED_WINDOW (widget));
  g_return_if_fail (allocation != NULL);
  
  zenith_theme_data = widget->style->engine_data;
  if (zenith_theme_data)
  {
    sb_place  = zenith_theme_data->sb_place;
    sb_indent = zenith_theme_data->sb_indent;
  }
  else
  {
    g_warning ("zenith_scrolled_window_size_allocate: Theme Data is NULL");
    sb_place = ZENITH_THEME_DEFAULT_SB_PLACE;
    sb_indent = ZENITH_THEME_DEFAULT_SB_INDENT;
  }
    
  switch (sb_place) {
  case ZENITH_SB_PLACE_TOP_LEFT:
    corner = GTK_CORNER_BOTTOM_RIGHT;
    break;
    
  case ZENITH_SB_PLACE_BOTTOM_LEFT:
    corner = GTK_CORNER_TOP_RIGHT;
    break;
    
  case ZENITH_SB_PLACE_TOP_RIGHT:
    corner = GTK_CORNER_BOTTOM_LEFT;
    break;
    
  case ZENITH_SB_PLACE_BOTTOM_RIGHT:
    corner = GTK_CORNER_TOP_LEFT;
    break;
  }
  gtk_scrolled_window_set_placement (GTK_SCROLLED_WINDOW (widget), corner);
  
  scrolled_window = GTK_SCROLLED_WINDOW (widget);
  bin = GTK_BIN (scrolled_window);
  
  widget->allocation = *allocation;
  
  if (scrolled_window->hscrollbar_policy == GTK_POLICY_ALWAYS)
    scrolled_window->hscrollbar_visible = TRUE;
  else if (scrolled_window->hscrollbar_policy == GTK_POLICY_NEVER)
    scrolled_window->hscrollbar_visible = FALSE;
  if (scrolled_window->vscrollbar_policy == GTK_POLICY_ALWAYS)
    scrolled_window->vscrollbar_visible = TRUE;
  else if (scrolled_window->vscrollbar_policy == GTK_POLICY_NEVER)
    scrolled_window->vscrollbar_visible = FALSE;
  
  if (bin->child && GTK_WIDGET_VISIBLE (bin->child))
  {
    gboolean previous_hvis;
    gboolean previous_vvis;
    guint count = 0;
    
    do
    {
      zenith_scrolled_window_relative_allocation (widget, &relative_allocation);
	  
      child_allocation.x = relative_allocation.x + allocation->x;
      child_allocation.y = relative_allocation.y + allocation->y;
      child_allocation.width = relative_allocation.width;
      child_allocation.height = relative_allocation.height;
      
      previous_hvis = scrolled_window->hscrollbar_visible;
      previous_vvis = scrolled_window->vscrollbar_visible;
      
      gtk_widget_size_allocate (bin->child, &child_allocation);
      
      /* If, after the first iteration, the hscrollbar and the
       * vscrollbar flip visiblity, then we need both.
       */
      if (count &&
	  previous_hvis != scrolled_window->hscrollbar_visible &&
	  previous_vvis != scrolled_window->vscrollbar_visible)
      {
	scrolled_window->hscrollbar_visible = TRUE;
	scrolled_window->vscrollbar_visible = TRUE;
	
	/* a new resize is already queued at this point,
	 * so we will immediatedly get reinvoked
	 */
	return;
      }
      
      count++;
    }
    while (previous_hvis != scrolled_window->hscrollbar_visible ||
	   previous_vvis != scrolled_window->vscrollbar_visible);
  }
  else
    zenith_scrolled_window_relative_allocation (widget, &relative_allocation);
  
  if (scrolled_window->hscrollbar_visible)
  {
    GtkRequisition hscrollbar_requisition;
    gtk_widget_get_child_requisition (scrolled_window->hscrollbar,
				      &hscrollbar_requisition);
    
    if (!GTK_WIDGET_VISIBLE (scrolled_window->hscrollbar))
      gtk_widget_show (scrolled_window->hscrollbar);
    
    child_allocation.x = relative_allocation.x;
    if (scrolled_window->window_placement == GTK_CORNER_TOP_LEFT ||
	scrolled_window->window_placement == GTK_CORNER_TOP_RIGHT)
      child_allocation.y = (relative_allocation.y +
			    relative_allocation.height +
			    SCROLLBAR_SPACING (scrolled_window));
    else
      child_allocation.y = GTK_CONTAINER (scrolled_window)->border_width;
    
    child_allocation.width = relative_allocation.width;
    child_allocation.height = hscrollbar_requisition.height;
    child_allocation.x += allocation->x;
    child_allocation.y += allocation->y;

    if (scrolled_window->vscrollbar_visible &&
	(sb_indent == ZENITH_SB_INDENT_HORIZONTAL))
    {
      if (ZENITH_SB_PLACE_LEFT (zenith_theme_data->sb_place))
      {
	child_allocation.x -= (SCROLLBAR_SPACING (scrolled_window) +
			       scrolled_window->vscrollbar->requisition.width);
	child_allocation.width += (SCROLLBAR_SPACING (scrolled_window) +
				   scrolled_window->vscrollbar->requisition.width);
      }
      else if (ZENITH_SB_PLACE_RIGHT (zenith_theme_data->sb_place))
	child_allocation.width += (SCROLLBAR_SPACING (scrolled_window) +
				   scrolled_window->vscrollbar->requisition.width);
    }

    gtk_widget_size_allocate (scrolled_window->hscrollbar, &child_allocation);
  }
  else if (GTK_WIDGET_VISIBLE (scrolled_window->hscrollbar))
    gtk_widget_hide (scrolled_window->hscrollbar);
  
  if (scrolled_window->vscrollbar_visible)
  {
    GtkRequisition vscrollbar_requisition;
    if (!GTK_WIDGET_VISIBLE (scrolled_window->vscrollbar))
      gtk_widget_show (scrolled_window->vscrollbar);
    
    gtk_widget_get_child_requisition (scrolled_window->vscrollbar,
				      &vscrollbar_requisition);
    
    if (scrolled_window->window_placement == GTK_CORNER_TOP_LEFT ||
	scrolled_window->window_placement == GTK_CORNER_BOTTOM_LEFT)
      child_allocation.x = (relative_allocation.x +
			    relative_allocation.width +
			    SCROLLBAR_SPACING (scrolled_window));
    else
      child_allocation.x = GTK_CONTAINER (scrolled_window)->border_width;
    
    child_allocation.y = relative_allocation.y;
    child_allocation.width = vscrollbar_requisition.width;
    child_allocation.height = relative_allocation.height;
    child_allocation.x += allocation->x;
    child_allocation.y += allocation->y;
    
    if (scrolled_window->hscrollbar_visible &&
	(sb_indent == ZENITH_SB_INDENT_VERTICAL))
    {
      if (ZENITH_SB_PLACE_TOP (zenith_theme_data->sb_place))
      {
	child_allocation.y -= (SCROLLBAR_SPACING (scrolled_window) +
			       scrolled_window->hscrollbar->requisition.height);
	child_allocation.height += (SCROLLBAR_SPACING (scrolled_window) +
				    scrolled_window->hscrollbar->requisition.height);
      }
      else if (ZENITH_SB_PLACE_BOTTOM (zenith_theme_data->sb_place))
	child_allocation.height += (SCROLLBAR_SPACING (scrolled_window) +
				    scrolled_window->hscrollbar->requisition.height);
    }
    
    gtk_widget_size_allocate (scrolled_window->vscrollbar, &child_allocation);
  }
  else if (GTK_WIDGET_VISIBLE (scrolled_window->vscrollbar))
    gtk_widget_hide (scrolled_window->vscrollbar);
}

/*
** Bring our theme to life.
*/
void
theme_init (GtkThemeEngine *engine)
{
  engine->parse_rc_style    = theme_parse_rc_style;
  engine->merge_rc_style    = theme_merge_rc_style;
  engine->rc_style_to_style = theme_rc_style_to_style;
  engine->duplicate_style   = theme_duplicate_style;
  engine->realize_style     = theme_realize_style;
  engine->unrealize_style   = theme_unrealize_style;
  engine->destroy_rc_style  = theme_destroy_rc_style;
  engine->destroy_style     = theme_destroy_style;
  engine->set_background    = NULL;
  
  /* Make scrollbar wider and move arrows according to */
  /* the style's theme data                             */
  {
    GtkRangeClass  *rangeclass;
    GtkWidgetClass *widgetclass;

    rangeclass = GTK_RANGE_CLASS (gtk_type_class (gtk_hscrollbar_get_type ()));
    widgetclass = GTK_WIDGET_CLASS (rangeclass);
    rangeclass->slider_width           = ZENITH_THEME_DEFAULT_SCROLLBAR_WIDTH;
    rangeclass->min_slider_size        = ZENITH_THEME_DEFAULT_SCROLLBAR_WIDTH;
    rangeclass->stepper_size           = ZENITH_THEME_DEFAULT_SCROLLBAR_WIDTH;
    rangeclass->stepper_slider_spacing = 0;

    rangeclass->motion         = zenith_range_hmotion;
    rangeclass->trough_click   = zenith_range_htrough_click;
    rangeclass->slider_update  = zenith_hscrollbar_slider_update;
    widgetclass->realize       = zenith_hscrollbar_realize;
    widgetclass->size_allocate = zenith_hscrollbar_size_allocate;

    rangeclass = GTK_RANGE_CLASS (gtk_type_class (gtk_vscrollbar_get_type ()));
    widgetclass = GTK_WIDGET_CLASS (rangeclass);
    rangeclass->slider_width           = ZENITH_THEME_DEFAULT_SCROLLBAR_WIDTH;
    rangeclass->min_slider_size        = ZENITH_THEME_DEFAULT_SCROLLBAR_WIDTH;
    rangeclass->stepper_size           = ZENITH_THEME_DEFAULT_SCROLLBAR_WIDTH;
    rangeclass->stepper_slider_spacing = 0;
    
    rangeclass->motion         = zenith_range_vmotion;
    rangeclass->trough_click   = zenith_range_vtrough_click;
    rangeclass->slider_update  = zenith_vscrollbar_slider_update;
    widgetclass->realize       = zenith_vscrollbar_realize;
    widgetclass->size_allocate = zenith_vscrollbar_size_allocate;
  }

  /* Set window placement according to the style's theme data */
  {
    GtkScrolledWindowClass *scrolledwindowclass;
    GtkWidgetClass         *widgetclass;

    scrolledwindowclass = GTK_SCROLLED_WINDOW_CLASS (gtk_type_class (gtk_scrolled_window_get_type ()));
    widgetclass = GTK_WIDGET_CLASS (scrolledwindowclass);
    scrolledwindowclass->scrollbar_spacing = 1;
    widgetclass->size_allocate             = zenith_scrolled_window_size_allocate;
  }

  /* Alter the size of radio and check buttons */
  {
    GtkCheckButtonClass *radiobuttonclass;
    GtkCheckButtonClass *checkbuttonclass;

    radiobuttonclass =
      GTK_CHECK_BUTTON_CLASS (gtk_type_class (gtk_radio_button_get_type ()));
    radiobuttonclass->indicator_size = ZENITH_THEME_DEFAULT_RADIO_CHECK_SIZE;
    
    checkbuttonclass =
      GTK_CHECK_BUTTON_CLASS (gtk_type_class (gtk_check_button_get_type ()));
    checkbuttonclass->indicator_size = ZENITH_THEME_DEFAULT_RADIO_CHECK_SIZE;
  }
}

void
theme_exit (void)
{
}

/*
** The following function will be called by GTK+ when the module
** is loaded and checks to see if we are compatible with the
** version of GTK+ that loads us.
*/
G_MODULE_EXPORT const gchar *
g_module_check_init (GModule *module);

const gchar *
g_module_check_init (GModule *module)
{
  return gtk_check_version (GTK_MAJOR_VERSION,
			    GTK_MINOR_VERSION,
			    GTK_MICRO_VERSION - GTK_INTERFACE_AGE);
}
